# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Copyright (C) 2019-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.


"""SegWit address functions.

Some of these functions were originally from
https://github.com/sipa/bech32/tree/master/ref/python,
with the following modifications:

* moved bech32 stuff into bech32.py
* type annotated python3
* avoided returning None or (None, None), throwing Exceptions instead
* detailed error messages and exteded safety checks
* check that bech32 addresses are not longer than 90 characters
  (as this is not enforced by bech32.b32decode anymore)
"""


from typing import Iterable, List, Optional, Tuple

from .alias import Octets, String
from .bech32 import b32decode, b32encode
from .exceptions import BTClibValueError
from .hashes import hash160_from_key
from .network import NETWORKS, network_from_key_value
from .to_pub_key import Key
from .utils import bytes_from_octets, sha256

# 0. bech32 facilities


def _convertbits(
    data: Iterable[int], frombits: int, tobits: int, pad: bool = True
) -> List[int]:
    "General power-of-2 base conversion."
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            raise BTClibValueError(f"invalid value in _convertbits: {value}")
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)

    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits:
        raise BTClibValueError("zero padding of more than 4 bits in 8-to-5 conversion")
    elif (acc << (tobits - bits)) & maxv:
        raise BTClibValueError("non-zero padding in 8-to-5 conversion")

    return ret


def _check_witness(witvers: int, witprog: bytes):
    length = len(witprog)
    if witvers == 0:
        if length not in (20, 32):
            err_msg = "invalid witness program length for witness v0: "
            err_msg += f"{length} instead of 20 or 32"
            raise BTClibValueError(err_msg)
    elif witvers < 0 or witvers > 16:
        err_msg = "invalid witness version: "
        err_msg += f"{witvers} not in 0..16"
        raise BTClibValueError(err_msg)
    else:
        if length < 2 or length > 40:
            err_msg = "invalid witness program length for witness v0: "
            err_msg += f"{length}, not in 2..40"
            raise BTClibValueError(err_msg)


# 1. Hash/WitnessProgram from pub_key/script_pub_key
# imported from the hashes module

# 2. bech32 address from WitnessProgram and vice versa


def _bech32_address_from_witness(hrp: str, wit_ver: int, wit_prg: Octets) -> str:
    wit_prg = bytes_from_octets(wit_prg)
    _check_witness(wit_ver, wit_prg)
    bytes_ = b32encode(hrp, [wit_ver] + _convertbits(wit_prg, 8, 5))
    return bytes_.decode("ascii")


def bech32_address_from_witness(
    wit_ver: int, wit_prg: Octets, network: str = "mainnet"
) -> str:
    "Encode a bech32 native SegWit address from the witness."

    hrp = NETWORKS[network].p2w
    return _bech32_address_from_witness(hrp, wit_ver, wit_prg)


def witness_from_bech32_address(b32addr: String) -> Tuple[int, bytes, str, bool]:
    "Return the witness from a bech32 native SegWit address."

    if isinstance(b32addr, str):
        b32addr = b32addr.strip()

    # the following check was originally in b32decode
    # but it does not pertain there
    if len(b32addr) > 90:
        raise BTClibValueError(f"invalid bech32 address length: {len(b32addr)} > 90")

    hrp, data = b32decode(b32addr)

    # check that it is a known SegWit address type
    network = network_from_key_value("p2w", hrp)

    if len(data) == 0:
        raise BTClibValueError(f"empty data in bech32 address: {b32addr!r}")

    witvers = data[0]
    witprog = _convertbits(data[1:], 5, 8, False)
    _check_witness(witvers, bytes(witprog))

    is_script_hash = witvers != 0 or len(witprog) != 20
    return witvers, bytes(witprog), network, is_script_hash


# 1.+2. = 3. bech32 address from pub_key/script_pub_key


def p2wpkh(key: Key, network: Optional[str] = None) -> str:
    "Return the p2wpkh bech32 address corresponding to a public key."
    compressed = True  # needed to force check on pub_key
    h160, network = hash160_from_key(key, network, compressed)
    return bech32_address_from_witness(0, h160, network)


def p2wsh(script_pub_key: Octets, network: str = "mainnet") -> str:
    "Return the p2wsh bech32 address corresponding to a script_pub_key."
    h256 = sha256(script_pub_key)
    return bech32_address_from_witness(0, h256, network)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""BIP32 derivation path and key origin.

A BIP 32 derivation path can be represented as:

- "m/44h/0'/1H/0/10" or "44h/0'/1H/0/10" string
- sequence of integer indexes (even a single int)
- bytes (multiples of 4-bytes index)
"""

from dataclasses import InitVar, dataclass, field
from io import SEEK_CUR
from typing import (
    Any,
    Collection,
    Dict,
    List,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from dataclasses_json import DataClassJsonMixin, config

from btclib.sec_point import point_from_octets

from . import var_bytes
from .alias import BinaryData
from .exceptions import BTClibValueError
from .utils import bytes_from_octets, bytesio_from_binarydata

# default hardening symbol among the possible ones: "h", "H", "'"
_HARDENING = "h"


def _int_from_index_str(s: str) -> int:

    s.strip().lower()
    hardened = False
    if s[-1] in ("'", "h"):
        s = s[:-1]
        hardened = True

    index = int(s)
    if not 0 <= index < 0x80000000:
        raise BTClibValueError(f"invalid index: {index}")
    return index + (0x80000000 if hardened else 0)


def _str_from_index_int(i: int, hardening: str = _HARDENING) -> str:

    if hardening not in ("'", "h", "H"):
        raise BTClibValueError(f"invalid hardening symbol: {hardening}")
    if not 0 <= i <= 0xFFFFFFFF:
        raise BTClibValueError(f"invalid index: {i}")
    if i < 0x80000000:
        return str(i)
    return str(i - 0x80000000) + hardening


def _indexes_from_bip32_path_str(der_path: str, skip_m: bool = True) -> List[int]:

    steps = [x.strip().lower() for x in der_path.split("/")]
    if skip_m and steps[0] == "m":
        steps = steps[1:]

    indexes = [_int_from_index_str(s) for s in steps if s != ""]

    if len(indexes) > 255:
        err_msg = f"depth greater than 255: {len(indexes)}"
        raise BTClibValueError(err_msg)
    return indexes


BIP32DerPath = Union[str, Sequence[int], int, bytes]


def indexes_from_bip32_path(der_path: BIP32DerPath) -> List[int]:

    if isinstance(der_path, str):
        return _indexes_from_bip32_path_str(der_path)

    if isinstance(der_path, int):
        return [der_path]

    if isinstance(der_path, bytes):
        if len(der_path) % 4 != 0:
            err_msg = f"index are not a multiple of 4-bytes: {len(der_path)}"
            raise BTClibValueError(err_msg)
        return [
            int.from_bytes(der_path[n : n + 4], byteorder="little", signed=False)
            for n in range(0, len(der_path), 4)
        ]

    # Iterable[int]
    return [int(i) for i in der_path]


def _str_from_bip32_path(der_path: BIP32DerPath, hardening: str = _HARDENING) -> str:
    indexes = indexes_from_bip32_path(der_path)
    return "/".join(_str_from_index_int(i, hardening) for i in indexes)


def str_from_bip32_path(der_path: BIP32DerPath, hardening: str = _HARDENING) -> str:
    result = _str_from_bip32_path(der_path, hardening)
    return "m/" + result if result else "m"


def bytes_from_bip32_path(der_path: BIP32DerPath) -> bytes:
    indexes = indexes_from_bip32_path(der_path)
    result = [i.to_bytes(4, byteorder="little", signed=False) for i in indexes]
    return b"".join(result)


_BIP32KeyOrigin = TypeVar("_BIP32KeyOrigin", bound="BIP32KeyOrigin")


@dataclass(frozen=True)
class BIP32KeyOrigin(DataClassJsonMixin):
    fingerprint: bytes = field(
        metadata=config(
            field_name="master_fingerprint",
            encoder=lambda v: v.hex(),
            decoder=bytes.fromhex,  # use bytes_from_octets(v, 4)
        ),
    )
    der_path: Sequence[int] = field(
        metadata=config(
            field_name="path",
            encoder=str_from_bip32_path,
            decoder=indexes_from_bip32_path,
        ),
    )
    check_validity: InitVar[bool] = True

    def __post_init__(self, check_validity: bool) -> None:
        if check_validity:
            self.assert_valid()

    def __len__(self):
        return len(self.der_path)

    @property
    def description(self) -> str:

        fingerprint = self.fingerprint.hex()
        der_path = _str_from_bip32_path(self.der_path, hardening=_HARDENING)
        return fingerprint + "/" + der_path if der_path else fingerprint

    def assert_valid(self) -> None:
        if len(self.fingerprint) != 4:
            err_msg = f"invalid master fingerprint length: {len(self.fingerprint)}"
            raise BTClibValueError(err_msg)
        if len(self) > 255:
            raise BTClibValueError(f"invalid der_path size: {len(self.der_path)}")
        if any(not 0 <= i <= 0xFFFFFFFF for i in self.der_path):
            raise BTClibValueError("invalid der_path element")

    def serialize(self, check_validity: bool = True) -> bytes:

        if check_validity:
            self.assert_valid()

        return self.fingerprint + bytes_from_bip32_path(self.der_path)

    @classmethod
    def deserialize(
        cls: Type[_BIP32KeyOrigin], data: bytes, check_validity: bool = True
    ) -> _BIP32KeyOrigin:
        "Return a BIP32KeyOrigin by parsing binary data."

        fingerprint = data[:4]
        der_path = indexes_from_bip32_path(data[4:])

        return cls(fingerprint, der_path, check_validity)

    @classmethod
    def from_description(
        cls: Type[_BIP32KeyOrigin], data: str, check_validity: bool = True
    ) -> _BIP32KeyOrigin:
        data = data.strip()
        fingerprint = bytes.fromhex(data[:8])
        der_path = _indexes_from_bip32_path_str(data[9:], skip_m=False)

        return cls(fingerprint, der_path, check_validity)


_BIP32KeyPath = TypeVar("_BIP32KeyPath", bound="BIP32KeyPath")


@dataclass(frozen=True)
class BIP32KeyPath(DataClassJsonMixin):
    pub_key: bytes = field(
        metadata=config(
            encoder=lambda v: v.hex(),
            decoder=bytes.fromhex,
        ),
    )
    key_origin: BIP32KeyOrigin
    check_validity: InitVar[bool] = True

    def __post_init__(self, check_validity: bool) -> None:
        if check_validity:
            self.assert_valid()

    def __len__(self):
        return len(self.key_origin)

    def assert_valid(self) -> None:
        # check that self.pub_key is a valid SEC key
        point_from_octets(self.pub_key)
        self.key_origin.assert_valid()

    def serialize(self, check_validity: bool = True) -> bytes:

        if check_validity:
            self.assert_valid()

        bytes_ = var_bytes.serialize(self.pub_key)
        temp = self.key_origin.serialize()
        bytes_ += var_bytes.serialize(temp)
        return var_bytes.serialize(bytes_)

    @classmethod
    def deserialize(
        cls: Type[_BIP32KeyPath], data: BinaryData, check_validity: bool = True
    ) -> _BIP32KeyPath:
        "Return a BIP32KeyPath by parsing binary data."

        bytes_ = var_bytes.deserialize(data)
        stream = bytesio_from_binarydata(bytes_)
        pub_key = var_bytes.deserialize(stream)
        der_path_bytes = var_bytes.deserialize(stream)
        key_origin = BIP32KeyOrigin.deserialize(der_path_bytes, check_validity)

        return cls(pub_key, key_origin, check_validity)


def _encode_hd_key_paths(
    dictionary: Dict[bytes, BIP32KeyOrigin]
) -> List[Dict[str, Union[str, BIP32KeyOrigin]]]:
    "Return the json representation of the dataclass element."

    return [
        {
            "pub_key": pub_key.hex(),
            "key_origin": key_origin,
        }
        for pub_key, key_origin in dictionary.items()
    ]


def _decode_hd_key_path(new_element: Dict[str, Any]) -> Tuple[bytes, BIP32KeyOrigin]:
    fingerprint = bytes_from_octets(new_element["key_origin"]["master_fingerprint"], 4)
    der_path = indexes_from_bip32_path(new_element["key_origin"]["path"])
    k = bytes_from_octets(new_element["pub_key"], [33, 65, 78])
    return k, BIP32KeyOrigin(fingerprint, der_path)


def _decode_hd_key_paths(
    list_of_dict: List[Dict[str, Collection[str]]]
) -> Dict[bytes, BIP32KeyOrigin]:
    "Return the dataclass element from its json representation."

    return dict([_decode_hd_key_path(item) for item in list_of_dict])


def _serialize_hd_key_paths(
    type_: bytes, dictionary: Dict[bytes, BIP32KeyOrigin]
) -> bytes:
    "Return the binary representation of the dataclass element."

    if len(type_) != 1:
        raise BTClibValueError("invalid type marker")

    return b"".join(
        [
            var_bytes.serialize(type_ + k) + var_bytes.serialize(v.serialize())
            for k, v in sorted(dictionary.items())
        ]
    )


def _assert_valid_hd_key_paths(hd_key_paths: Dict[bytes, BIP32KeyOrigin]) -> None:
    "Raise an exception if the dataclass element is not valid."

    allowed_lengths = (78, 33, 65)
    for pub_key, key_origin in hd_key_paths.items():
        # test vector 6 contains an invalid pubkey
        # point_from_pub_key(pub_key)
        if len(pub_key) not in allowed_lengths:
            err_msg = f"invalid public key length: {len(pub_key)}"
            raise BTClibValueError(err_msg)
        key_origin.assert_valid()


_BIP32KeyPaths = TypeVar("_BIP32KeyPaths", bound="BIP32KeyPaths")


@dataclass
class BIP32KeyPaths(DataClassJsonMixin):
    hd_key_paths: Dict[bytes, BIP32KeyOrigin] = field(
        default_factory=dict,
        metadata=config(
            field_name="bip32_derivs",
            encoder=_encode_hd_key_paths,
            decoder=_decode_hd_key_paths,
        ),
    )
    check_validity: InitVar[bool] = True

    def __post_init__(self, check_validity: bool) -> None:
        self.hd_key_paths = dict(sorted(self.hd_key_paths.items()))
        if check_validity:
            self.assert_valid()

    def __len__(self):
        return len(self.hd_key_paths)

    def assert_valid(self) -> None:
        _assert_valid_hd_key_paths(self.hd_key_paths)

    def serialize(self, type_: bytes, check_validity: bool = True) -> bytes:

        if check_validity:
            self.assert_valid()

        if len(type_) != 1:
            raise BTClibValueError("invalid type marker")

        return _serialize_hd_key_paths(type_, self.hd_key_paths)

    @classmethod
    def deserialize(
        cls: Type[_BIP32KeyPaths],
        data: BinaryData,
        type_: bytes,
        check_validity: bool = True,
    ) -> _BIP32KeyPaths:
        "Return a BIP32KeyPaths by parsing binary data."

        if len(type_) != 1:
            raise BTClibValueError("invalid type marker")

        stream = bytesio_from_binarydata(data)
        hd_key_paths: Dict[bytes, BIP32KeyOrigin] = {}
        while stream.read(1):
            stream.seek(-1, SEEK_CUR)
            prefixed_pubkey = var_bytes.deserialize(stream)
            if prefixed_pubkey and prefixed_pubkey[:1] == type_:
                pubkey = prefixed_pubkey[1:]
                if pubkey in hd_key_paths:
                    raise BTClibValueError("duplicate pubkey")
                key_origin_bytes = var_bytes.deserialize(stream)
                key_origin = BIP32KeyOrigin.deserialize(key_origin_bytes)
                hd_key_paths[pubkey] = key_origin

        return cls(hd_key_paths, check_validity)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Assorted conversion utilities.

Most conversions from SEC 1 v.2 2.3 are included.

https://www.secg.org/sec1-v2.pdf
"""

import hashlib
from collections.abc import Iterable as IterableCollection
from io import BytesIO
from typing import Callable, Iterable, List, Optional, Union

from .alias import BinaryData, Integer, Octets
from .exceptions import BTClibTypeError, BTClibValueError

# hexstr_from_bytes is not needed!!
# def hexstr_from_bytes(byte_str: bytes) -> str:
#    return byte_str.hex()


def sha256(octets: Octets) -> bytes:
    "Return the SHA256(*) of the input octet sequence."

    octets = bytes_from_octets(octets)
    return hashlib.sha256(octets).digest()


def hash160(octets: Octets) -> bytes:
    "Return the HASH160=RIPEMD160(SHA256) of the input octet sequence."

    t = sha256(octets)
    return hashlib.new("ripemd160", t).digest()


def hash256(octets: Octets) -> bytes:
    "Return the SHA256(SHA256(*)) of the input octet sequence."

    t = sha256(octets)
    return hashlib.sha256(t).digest()


def merkle_root(data: List[bytes], hf: Callable[[Union[bytes, str]], bytes]) -> bytes:
    """Return the Merkel tree root of a list of binary hashes.

    The Merkel tree is a binary tree constructed
    with the provided list of binary data as bottom level,
    then recursively going up one level
    by hashing every hash value pair in the current level,
    until a single value (root) is obtained.
    """

    data = [hf(item) for item in data]
    while len(data) != 1:
        parent_level = []
        if len(data) % 2:
            data.append(data[-1])
        for i in range(0, len(data), 2):
            parent = hf(data[i] + data[i + 1])
            parent_level.append(parent)
        data = parent_level[:]
    return data[0]


NoneOneOrMoreInt = Optional[Union[int, Iterable[int]]]


def bytes_from_octets(octets: Octets, out_size: NoneOneOrMoreInt = None) -> bytes:
    """Return bytes from a hex-string, stripping leading/trailing spaces.

    If the input is not a string, then it goes untouched.
    Optionally, it also ensures required output size.
    """

    if isinstance(octets, str):  # hex string
        octets = bytes.fromhex(octets)

    if (
        out_size is None
        or isinstance(out_size, int)
        and len(octets) == out_size
        or isinstance(out_size, IterableCollection)
        and len(octets) in out_size
    ):
        return octets

    m = f"invalid size: {len(octets)} bytes instead of {out_size}"
    raise BTClibValueError(m)


def bytesio_from_binarydata(stream: BinaryData) -> BytesIO:
    """Return a BytesIO stream object from BinaryIO or Octets.

    If the input is not Octets (i.e. str or bytes),
    then it goes untouched.
    """

    if isinstance(stream, str):  # hex string
        stream = bytes_from_octets(stream)

    if isinstance(stream, bytes):
        stream = BytesIO(stream)

    return stream


def int_from_bits(octets: Octets, nlen: int) -> int:
    """Return the leftmost nlen bits.

    Take as input a sequence of blen bits and calculate a
    non-negative integer i that is less than 2^nlen according to
    SEC 1 v.2 section 4.1.3 (5).
    Note that an additional reduction modulo n would be required
    to ensure that 0 < i < n.

    int_from_bits is not the reverse of i.to_bytes, even
    for input sequences of length nlen: i.to_bytes will add some
    bits on the left, while int_from_bits will discard some bits on the
    right. i.to_bytes is the reverse of int_from_bits only when
    nlen is a multiple of 8 and bit sequences already have length nlen.
    See https://tools.ietf.org/html/rfc6979#section-2.3.5.
    """

    octets = bytes_from_octets(octets)
    i = int.from_bytes(octets, byteorder="big", signed=False)

    blen = len(octets) * 8  # bits
    n = (blen - nlen) if blen >= nlen else 0
    return i >> n


def int_from_integer(i: Integer) -> int:
    """Return an int from many possible integer representations.

    Allowed integer representations are:

    * 3735928559
    * -3735928559
    * "0xdeadbeef"
    * "-0xdeadbeef"
    * "deadbeef"
    * b'\xde\xad\xbe\xef'

    The binary representation is not allowed because there is no way to
    discriminate it from a valid hex-string
    (e.g. "0b11011110101011011011111011101111").
    """

    if isinstance(i, int):
        return i
    if isinstance(i, str):
        i = i.strip().lower()
        if i.startswith("0x") or i.startswith("-0x"):
            return int(i, 16)
        i = bytes.fromhex(i)
    if not isinstance(i, bytes):
        raise BTClibTypeError("not an Integer=Union[int, str, bytes]")
    return int.from_bytes(i, "big", signed=False)


def hex_string(i: Integer) -> str:
    """Return a hex-string from many positive integer representations.

    Negative integers are not allowed.

    The resulting hex-string has an even number of hex-digits and
    includes a space every four bytes (i.e. every eight hex-digits).
    """

    int_ = int_from_integer(i)
    if int_ < 0:
        raise BTClibValueError(f"negative integer: {int_}")
    a_str = hex(int_)[2:]
    if len(a_str) % 2 != 0:
        a_str = "0" + a_str

    indx = list(reversed(range(len(a_str), 0, -8)))
    lresult = [(a_str[max(0, i - 8) : i]) for i in indx]
    result = " ".join(lresult)
    return result.upper()


def ensure_is_power_of_two(n: int, var_name: str = None) -> None:
    # http://www.graphics.stanford.edu/~seander/bithacks.html
    if n & (n - 1) != 0:
        raise BTClibValueError(f"{var_name}: {n} (must be a power of two)")



#!/usr/bin/env python3

# Copyright (C) 2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.network` module."

import json
from os import path

import pytest

from btclib.network import (
    NETWORKS,
    Network,
    curve_from_xkeyversion,
    network_from_xkeyversion,
    xprvversions_from_network,
    xpubversions_from_network,
)


def test_curve_from_xkeyversion() -> None:
    for net in NETWORKS:
        all_versions = xpubversions_from_network(net) + xprvversions_from_network(net)
        for version in all_versions:
            # unfortunately 'regtest' shares same versions with 'testnet'
            if net != "regtest":
                assert net == network_from_xkeyversion(version)
            assert NETWORKS[net].curve == curve_from_xkeyversion(version)


def test_space_and_caps() -> None:
    net = " MainNet "
    assert xpubversions_from_network(net), f"unknown network: {net}"

    with pytest.raises(KeyError):
        net = " MainNet2 "
        xpubversions_from_network(net)


def test_numbers_of_networks() -> None:
    assert len(NETWORKS) == 3


def test_dataclasses_json_dict() -> None:
    for network_name, net in NETWORKS.items():
        net2 = Network.from_json(net.to_json())
        assert net2 == net
        net2 = Network.from_dict(net.to_dict())
        assert net2 == net

        datadir = path.join(path.dirname(__file__), "generated_files")
        filename = path.join(datadir, network_name + ".json")
        with open(filename, "w") as file_:
            json.dump(NETWORKS[network_name].to_dict(), file_, indent=4)



#!/usr/bin/env python3

# Copyright (C) 2020-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Partially Signed Bitcoin Transaction Input.

https://github.com/bitcoin/bips/blob/master/bip-0174.mediawiki
"""

from dataclasses import InitVar, dataclass, field
from typing import Dict, Optional, Type, TypeVar

from dataclasses_json import DataClassJsonMixin, config

from . import dsa, sec_point
from .bip32_path import BIP32KeyOrigin
from .exceptions import BTClibTypeError, BTClibValueError
from .psbt_out import (
    _assert_valid_hd_key_paths,
    _assert_valid_redeem_script,
    _assert_valid_unknown,
    _assert_valid_witness_script,
    _decode_dict_bytes_bytes,
    _decode_hd_key_paths,
    _deserialize_bytes,
    _encode_dict_bytes_bytes,
    _encode_hd_key_paths,
    _serialize_bytes,
    _serialize_dict_bytes_bytes,
    _serialize_hd_key_paths,
)
from .sign_hash import assert_valid_hash_type
from .tx import Tx
from .tx_out import TxOut
from .witness import Witness

PSBT_IN_NON_WITNESS_UTXO = b"\x00"
PSBT_IN_WITNESS_UTXO = b"\x01"
PSBT_IN_PARTIAL_SIG = b"\x02"
PSBT_IN_SIG_HASH_TYPE = b"\x03"
PSBT_IN_REDEEM_SCRIPT = b"\x04"
PSBT_IN_WITNESS_SCRIPT = b"\x05"
PSBT_IN_BIP32_DERIVATION = b"\x06"
PSBT_IN_FINAL_SCRIPTSIG = b"\x07"
PSBT_IN_FINAL_SCRIPTWITNESS = b"\x08"
# TODO: add support for the following
# PSBT_IN_RIPEMD160 = b"\0x0a"
# PSBT_IN_SHA256 = b"\0x0b"
# PSBT_IN_HASH160 = b"\0x0c"
# PSBT_IN_HASH256 = b"\0x0d"

# 0xfc is reserved for proprietary
# explicit code support for proprietary (and por) is unnecessary
# see https://github.com/bitcoin/bips/pull/1038
# PSBT_IN_PROPRIETARY = b"\xfc"


def _deserialize_tx(k: bytes, v: bytes, type_: str) -> Tx:
    "Return the dataclass element from its binary representation."

    if len(k) != 1:
        err_msg = f"invalid {type_} key length: {len(k)}"
        raise BTClibValueError(err_msg)
    return Tx.deserialize(v)


def _deserialize_witness_utxo(k: bytes, v: bytes) -> TxOut:
    "Return the dataclass element from its binary representation."

    if len(k) != 1:
        err_msg = f"invalid witness-utxo key length: {len(k)}"
        raise BTClibValueError(err_msg)
    return TxOut.deserialize(v)


def _assert_valid_partial_sigs(partial_sigs: Dict[bytes, bytes]) -> None:
    "Raise an exception if the dataclass element is not valid."

    for pub_key, sig in partial_sigs.items():
        try:
            # pub_key must be a valid secp256k1 Point in SEC representation
            sec_point.point_from_octets(pub_key)
        except BTClibValueError as e:
            err_msg = "invalid partial signature pub_key: {pub_key!r}"
            raise BTClibValueError(err_msg) from e
        try:
            dsa.Sig.deserialize(sig)
        except BTClibValueError as e:
            err_msg = f"invalid partial signature: {sig!r}"
            raise BTClibValueError(err_msg) from e
        # TODO should we check that pub_key is recoverable from sig?


def _deserialize_int(k: bytes, v: bytes, type_: str) -> int:
    "Return the dataclass element from its binary representation."

    if len(k) != 1:
        err_msg = f"invalid {type_} key length: {len(k)}"
        raise BTClibValueError(err_msg)
    return int.from_bytes(v, byteorder="little", signed=False)


def _assert_valid_final_script_sig(final_script_sig: bytes) -> None:
    if not isinstance(final_script_sig, bytes):
        raise BTClibTypeError("invalid final script_sig")


def _deserialize_final_script_witness(k: bytes, v: bytes) -> Witness:
    "Return the dataclass element from its binary representation."

    if len(k) != 1:
        err_msg = f"invalid final script witness key length: {len(k)}"
        raise BTClibValueError(err_msg)
    return Witness.deserialize(v)


_PsbtIn = TypeVar("_PsbtIn", bound="PsbtIn")


@dataclass
class PsbtIn(DataClassJsonMixin):
    non_witness_utxo: Optional[Tx] = None
    witness_utxo: Optional[TxOut] = None
    partial_sigs: Dict[bytes, bytes] = field(
        default_factory=dict,
        metadata=config(
            field_name="partial_signatures",
            encoder=_encode_dict_bytes_bytes,
            decoder=_decode_dict_bytes_bytes,
        ),
    )
    sig_hash_type: Optional[int] = field(
        default=None, metadata=config(field_name="sign_hash")
    )
    redeem_script: bytes = field(
        default=b"", metadata=config(encoder=lambda v: v.hex(), decoder=bytes.fromhex)
    )
    witness_script: bytes = field(
        default=b"", metadata=config(encoder=lambda v: v.hex(), decoder=bytes.fromhex)
    )
    hd_key_paths: Dict[bytes, BIP32KeyOrigin] = field(
        default_factory=dict,
        metadata=config(
            field_name="bip32_derivs",
            encoder=_encode_hd_key_paths,
            decoder=_decode_hd_key_paths,
        ),
    )
    final_script_sig: bytes = field(
        default=b"", metadata=config(encoder=lambda v: v.hex(), decoder=bytes.fromhex)
    )
    final_script_witness: Witness = Witness()
    unknown: Dict[bytes, bytes] = field(
        default_factory=dict,
        metadata=config(
            encoder=_encode_dict_bytes_bytes, decoder=_decode_dict_bytes_bytes
        ),
    )
    check_validity: InitVar[bool] = True

    @property
    def sig_hash(self) -> int:
        "Return the sig_hash int for compatibility with PartiallySignedInput."
        return self.sig_hash_type or 0

    def __post_init__(self, check_validity: bool) -> None:
        self.unknown = dict(sorted(self.unknown.items()))
        self.hd_key_paths = dict(sorted(self.hd_key_paths.items()))
        if check_validity:
            self.assert_valid()

    def assert_valid(self) -> None:
        "Assert logical self-consistency."

        if self.non_witness_utxo:
            self.non_witness_utxo.assert_valid()

        if self.witness_utxo:
            self.witness_utxo.assert_valid()

        _assert_valid_partial_sigs(self.partial_sigs)

        if self.sig_hash_type:
            assert_valid_hash_type(self.sig_hash_type)

        _assert_valid_redeem_script(self.redeem_script)
        _assert_valid_witness_script(self.witness_script)
        _assert_valid_hd_key_paths(self.hd_key_paths)
        _assert_valid_final_script_sig(self.final_script_sig)
        self.final_script_witness.assert_valid()

        _assert_valid_unknown(self.unknown)

    def serialize(self, check_validity: bool = True) -> bytes:

        if check_validity:
            self.assert_valid()

        psbt_in_bin = b""

        if self.non_witness_utxo:
            temp = self.non_witness_utxo.serialize(include_witness=True)
            psbt_in_bin += _serialize_bytes(PSBT_IN_NON_WITNESS_UTXO, temp)

        if self.witness_utxo:
            psbt_in_bin += _serialize_bytes(
                PSBT_IN_WITNESS_UTXO, self.witness_utxo.serialize()
            )

        if not self.final_script_sig and not self.final_script_witness:

            if self.partial_sigs:
                psbt_in_bin += _serialize_dict_bytes_bytes(
                    PSBT_IN_PARTIAL_SIG, self.partial_sigs
                )

            if self.sig_hash_type:
                temp = self.sig_hash_type.to_bytes(4, byteorder="little", signed=False)
                psbt_in_bin += _serialize_bytes(PSBT_IN_SIG_HASH_TYPE, temp)

            if self.redeem_script:
                psbt_in_bin += _serialize_bytes(
                    PSBT_IN_REDEEM_SCRIPT, self.redeem_script
                )

            if self.witness_script:
                psbt_in_bin += _serialize_bytes(
                    PSBT_IN_WITNESS_SCRIPT, self.witness_script
                )

            if self.hd_key_paths:
                psbt_in_bin += _serialize_hd_key_paths(
                    PSBT_IN_BIP32_DERIVATION, self.hd_key_paths
                )

        if self.final_script_sig:
            psbt_in_bin += _serialize_bytes(
                PSBT_IN_FINAL_SCRIPTSIG, self.final_script_sig
            )

        if self.final_script_witness:
            temp = self.final_script_witness.serialize()
            psbt_in_bin += _serialize_bytes(PSBT_IN_FINAL_SCRIPTWITNESS, temp)

        if self.unknown:
            psbt_in_bin += _serialize_dict_bytes_bytes(b"", self.unknown)

        return psbt_in_bin

    @classmethod
    def deserialize(
        cls: Type[_PsbtIn], input_map: Dict[bytes, bytes], check_validity: bool = True
    ) -> _PsbtIn:
        "Return a PsbtIn by parsing binary data."

        # FIX deserialize must use BinaryData

        non_witness_utxo = None
        witness_utxo = None
        partial_sigs: Dict[bytes, bytes] = {}
        sig_hash_type = None
        redeem_script = b""
        witness_script = b""
        hd_key_paths: Dict[bytes, BIP32KeyOrigin] = {}
        final_script_sig = b""
        final_script_witness = Witness()
        unknown: Dict[bytes, bytes] = {}

        for k, v in input_map.items():
            if k[:1] == PSBT_IN_NON_WITNESS_UTXO:
                if non_witness_utxo:
                    raise BTClibValueError("duplicate PsbtIn non_witness_utxo")
                non_witness_utxo = _deserialize_tx(k, v, "non-witness utxo")
            elif k[:1] == PSBT_IN_WITNESS_UTXO:
                if witness_utxo:
                    raise BTClibValueError("duplicate PsbtIn witness_utxo")
                witness_utxo = _deserialize_witness_utxo(k, v)
            elif k[:1] == PSBT_IN_PARTIAL_SIG:
                if k[1:] in partial_sigs:
                    raise BTClibValueError("duplicate PsbtIn partial_sigs")
                partial_sigs[k[1:]] = v
            elif k[:1] == PSBT_IN_SIG_HASH_TYPE:
                if sig_hash_type:
                    raise BTClibValueError("duplicate PsbtIn sig_hash_type")
                sig_hash_type = _deserialize_int(k, v, "sign_hash type")
            elif k[:1] == PSBT_IN_REDEEM_SCRIPT:
                if redeem_script:
                    raise BTClibValueError("duplicate PsbtIn redeem_script")
                redeem_script = _deserialize_bytes(k, v, "redeem script")
            elif k[:1] == PSBT_IN_WITNESS_SCRIPT:
                if witness_script:
                    raise BTClibValueError("duplicate PsbtIn witness_script")
                witness_script = _deserialize_bytes(k, v, "witness script")
            elif k[:1] == PSBT_IN_BIP32_DERIVATION:
                if k[1:] in hd_key_paths:
                    raise BTClibValueError("duplicate pub_key in PsbtIn hd_key_path")
                hd_key_paths[k[1:]] = BIP32KeyOrigin.deserialize(v)
            elif k[:1] == PSBT_IN_FINAL_SCRIPTSIG:
                if final_script_sig:
                    raise BTClibValueError("duplicate PsbtIn final_script_sig")
                final_script_sig = _deserialize_bytes(k, v, "final script_sig")
            elif k[:1] == PSBT_IN_FINAL_SCRIPTWITNESS:
                if final_script_witness:
                    raise BTClibValueError("duplicate PsbtIn final_script_witness")
                final_script_witness = _deserialize_final_script_witness(k, v)
            else:  # unknown
                if k in unknown:
                    raise BTClibValueError("duplicate PsbtIn unknown")
                unknown[k] = v

        return cls(
            non_witness_utxo,
            witness_utxo,
            partial_sigs,
            sig_hash_type,
            redeem_script,
            witness_script,
            hd_key_paths,
            final_script_sig,
            final_script_witness,
            unknown,
            check_validity,
        )



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Bitcoin Script.

https://en.bitcoin.it/wiki/Script

Scripts are represented by List[ScriptToken], where ScriptToken = Union[int, str, bytes]:

* int [-1, 16] are shorcuts for 'OP_1NEGATE', 'OP_0' - 'OP_16'
* str are for opcodes (e.g. 'OP_HASH160') or hexstring data
* bytes are for data (but integers are often casted to int)
"""

from typing import List, Sequence, Union

from .alias import BinaryData, Octets
from .exceptions import BTClibTypeError, BTClibValueError
from .utils import bytes_from_octets, bytesio_from_binarydata

OP_CODES = {
    # Constants
    "OP_0": b"\x00",
    "OP_FALSE": b"\x00",
    "OP_PUSHDATA1": b"\x4c",
    "OP_PUSHDATA2": b"\x4d",
    "OP_PUSHDATA4": b"\x4e",
    "OP_1NEGATE": b"\x4f",
    "OP_RESERVED": b"\x50",
    "OP_1": b"\x51",
    "OP_TRUE": b"\x51",
    "OP_2": b"\x52",
    "OP_3": b"\x53",
    "OP_4": b"\x54",
    "OP_5": b"\x55",
    "OP_6": b"\x56",
    "OP_7": b"\x57",
    "OP_8": b"\x58",
    "OP_9": b"\x59",
    "OP_10": b"\x5a",
    "OP_11": b"\x5b",
    "OP_12": b"\x5c",
    "OP_13": b"\x5d",
    "OP_14": b"\x5e",
    "OP_15": b"\x5f",
    "OP_16": b"\x60",
    # Flow control
    "OP_NOP": b"\x61",
    "OP_VER": b"\x62",
    "OP_IF": b"\x63",
    "OP_NOTIF": b"\x64",
    "OP_VERIF": b"\x65",
    "OP_VERNOTIF": b"\x66",
    "OP_ELSE": b"\x67",
    "OP_ENDIF": b"\x68",
    "OP_VERIFY": b"\x69",
    "OP_RETURN": b"\x6a",
    # Stack
    "OP_TOALTSTACK": b"\x6b",
    "OP_FROMALTSTACK": b"\x6c",
    "OP_IFDUP": b"\x73",
    "OP_DEPTH": b"\x74",
    "OP_DROP": b"\x75",
    "OP_DUP": b"\x76",
    "OP_NIP": b"\x77",
    "OP_OVER": b"\x78",
    "OP_PICK": b"\x79",
    "OP_ROLL": b"\x7a",
    "OP_ROT": b"\x7b",
    "OP_SWAP": b"\x7c",
    "OP_TUCK": b"\x7d",
    "OP_2DROP": b"\x6d",
    "OP_2DUP": b"\x6e",
    "OP_3DUP": b"\x6f",
    "OP_2OVER": b"\x70",
    "OP_2ROT": b"\x71",
    "OP_2SWAP": b"\x72",
    # Splice
    "OP_SIZE": b"\x82",
    # Bitwise logic
    "OP_EQUAL": b"\x87",
    "OP_EQUALVERIFY": b"\x88",
    "OP_RESERVED1": b"\x89",
    "OP_RESERVED2": b"\x8a",
    # Arithmetic
    "OP_1ADD": b"\x8b",  # without OP_, 1ADD would be a number
    "OP_1SUB": b"\x8c",
    "OP_NEGATE": b"\x8f",
    "OP_ABS": b"\x90",
    "OP_NOT": b"\x91",
    "OP_0NOTEQUAL": b"\x92",
    "OP_ADD": b"\x93",
    "OP_SUB": b"\x94",
    "OP_BOOLAND": b"\x9a",
    "OP_BOOLOR": b"\x9b",
    "OP_NUMEQUAL": b"\x9c",
    "OP_NUMEQUALVERIFY": b"\x9d",
    "OP_NUMNOTEQUAL": b"\x9e",
    "OP_LESSTHAN": b"\x9f",
    "OP_GREATERTHAN": b"\xa0",
    "OP_LESSTHANOREQUAL": b"\xa1",
    "OP_GREATERTHANOREQUAL": b"\xa2",
    "OP_MIN": b"\xa3",
    "OP_MAX": b"\xa4",
    "OP_WITHIN": b"\xa5",
    # Crypto
    "OP_RIPEMD160": b"\xa6",
    "OP_SHA1": b"\xa7",
    "OP_SHA256": b"\xa8",
    "OP_HASH160": b"\xa9",
    "OP_HASH256": b"\xaa",
    "OP_CODESEPARATOR": b"\xab",
    "OP_CHECKSIG": b"\xac",
    "OP_CHECKSIGVERIFY": b"\xad",
    "OP_CHECKMULTISIG": b"\xae",
    "OP_CHECKMULTISIGVERIFY": b"\xaf",
    # Locktime
    "OP_NOP2": b"\xb1",
    "OP_CHECKLOCKTIMEVERIFY": b"\xb1",
    "OP_NOP3": b"\xb2",
    "OP_CHECKSEQUENCEVERIFY": b"\xb2",
    # Reserved words
    "OP_NOP1": b"\xb0",
    "OP_NOP4": b"\xb3",
    "OP_NOP5": b"\xb4",
    "OP_NOP6": b"\xb5",
    "OP_NOP7": b"\xb6",
    "OP_NOP8": b"\xb7",
    "OP_NOP9": b"\xb8",
    "OP_NOP10": b"\xb9",
}

OP_CODE_NAMES = {
    0: "OP_0",
    76: "OP_PUSHDATA1",
    77: "OP_PUSHDATA2",
    78: "OP_PUSHDATA4",
    79: "OP_1NEGATE",
    80: "OP_RESERVED",
    81: "OP_1",
    82: "OP_2",
    83: "OP_3",
    84: "OP_4",
    85: "OP_5",
    86: "OP_6",
    87: "OP_7",
    88: "OP_8",
    89: "OP_9",
    90: "OP_10",
    91: "OP_11",
    92: "OP_12",
    93: "OP_13",
    94: "OP_14",
    95: "OP_15",
    96: "OP_16",
    97: "OP_NOP",
    98: "OP_VER",
    99: "OP_IF",
    100: "OP_NOTIF",
    101: "OP_VERIF",
    102: "OP_VERNOTIF",
    103: "OP_ELSE",
    104: "OP_ENDIF",
    105: "OP_VERIFY",
    106: "OP_RETURN",
    107: "OP_TOALTSTACK",
    108: "OP_FROMALTSTACK",
    109: "OP_2DROP",
    110: "OP_2DUP",
    111: "OP_3DUP",
    112: "OP_2OVER",
    113: "OP_2ROT",
    114: "OP_2SWAP",
    115: "OP_IFDUP",
    116: "OP_DEPTH",
    117: "OP_DROP",
    118: "OP_DUP",
    119: "OP_NIP",
    120: "OP_OVER",
    121: "OP_PICK",
    122: "OP_ROLL",
    123: "OP_ROT",
    124: "OP_SWAP",
    125: "OP_TUCK",
    130: "OP_SIZE",
    135: "OP_EQUAL",
    136: "OP_EQUALVERIFY",
    137: "OP_RESERVED1",
    138: "OP_RESERVED2",
    139: "OP_1ADD",
    140: "OP_1SUB",
    143: "OP_NEGATE",
    144: "OP_ABS",
    145: "OP_NOT",
    146: "OP_0NOTEQUAL",
    147: "OP_ADD",
    148: "OP_SUB",
    154: "OP_BOOLAND",
    155: "OP_BOOLOR",
    156: "OP_NUMEQUAL",
    157: "OP_NUMEQUALVERIFY",
    158: "OP_NUMNOTEQUAL",
    159: "OP_LESSTHAN",
    160: "OP_GREATERTHAN",
    161: "OP_LESSTHANOREQUAL",
    162: "OP_GREATERTHANOREQUAL",
    163: "OP_MIN",
    164: "OP_MAX",
    165: "OP_WITHIN",
    166: "OP_RIPEMD160",
    167: "OP_SHA1",
    168: "OP_SHA256",
    169: "OP_HASH160",
    170: "OP_HASH256",
    171: "OP_CODESEPARATOR",
    172: "OP_CHECKSIG",
    173: "OP_CHECKSIGVERIFY",
    174: "OP_CHECKMULTISIG",
    175: "OP_CHECKMULTISIGVERIFY",
    176: "OP_NOP1",
    177: "OP_CHECKLOCKTIMEVERIFY",
    178: "OP_CHECKSEQUENCEVERIFY",
    179: "OP_NOP4",
    180: "OP_NOP5",
    181: "OP_NOP6",
    182: "OP_NOP7",
    183: "OP_NOP8",
    184: "OP_NOP9",
    185: "OP_NOP10",
}


def _op_pushdata(data: Octets) -> bytes:
    """Convert to canonical push: OP_PUSHDATA (if needed) | length | data.

    According to standardness rules (BIP-62) the
    minimum possible PUSHDATA operator must be used.
    Byte vectors on the stack are not allowed to be more than 520 bytes long.
    """

    data = bytes_from_octets(data)
    r = b""
    length = len(data)
    if length < 76:  # 1-byte-length
        r += length.to_bytes(1, byteorder="little", signed=False)
    elif length < 256:  # OP_PUSHDATA1 | 1-byte-length
        r += OP_CODES["OP_PUSHDATA1"]
        r += length.to_bytes(1, byteorder="little", signed=False)
    elif length < 521:  # OP_PUSHDATA2 | 2-byte-length
        r += OP_CODES["OP_PUSHDATA2"]
        r += length.to_bytes(2, byteorder="little", signed=False)
    else:
        # because of the 520 bytes limit
        # there is no need to use OP_PUSHDATA4
        # r += OP_CODES['OP_PUSHDATA4']
        # r += length.to_bytes(4, byteorder="little", signed=False)
        raise BTClibValueError(f"Too many bytes for OP_PUSHDATA: {length}")
    r += data
    return r


def _op_int(token: int) -> bytes:
    # Short 1-byte opcodes exist
    # to push numbers in [-1, 16]
    if 0 <= token <= 16:
        return OP_CODES["OP_" + str(token)]
    if token == -1:
        return OP_CODES["OP_1NEGATE"]
    # Pushing any other number requires an
    # explicit push operation of its bytes encoding
    # FIXME: negative numbers?

    v = token
    # Convert number to bitcoin-specific little endian format
    # We need v.bit_length() bits, plus a sign bit for every nonzero number.
    n_bits = v.bit_length() + (v != 0)
    # The number of bytes for that is:
    n_bytes = (n_bits + 7) // 8
    # Convert number to absolute value + sign in top bit.
    encoded_v = 0 if v == 0 else abs(v) | ((v < 0) << (n_bytes * 8 - 1))
    # Serialize to bytes
    data = encoded_v.to_bytes(n_bytes, byteorder="little", signed=False)
    return _op_pushdata(data)


def _op_str(token: str) -> bytes:
    token = token.strip().upper()
    if token in OP_CODES:
        return OP_CODES[token]
    try:
        data = bytes.fromhex(token)
    except ValueError as e:
        raise BTClibValueError(f"invalid string token: {token}") from e
    return _op_pushdata(data)


# the integers [0-16] are shorcuts for 'OP_0'-'OP_16'
# the integer -1 is a shorcut for 'OP_1NEGATE'
# other integers are bytes encoded (require push operation)
# ascii str are for opcodes (e.g. 'OP_HASH160')
# Octets are for data to be pushed
ScriptToken = Union[int, str, bytes]


def serialize(script: Sequence[ScriptToken]) -> bytes:
    r = b""
    for token in script:
        if isinstance(token, int):
            r += _op_int(token)
        elif isinstance(token, str):
            r += _op_str(token)
        elif isinstance(token, bytes):
            r += _op_pushdata(token)
        else:
            raise BTClibTypeError(f"Unmanaged {type(token)} token type")
    return r


# Bitcoin script expressed as List[ScriptToken]
# e.g. [OP_HASH160, script_h160, OP_EQUAL]
# or Octets of its byte-encoded representation
Script = List[ScriptToken]


def deserialize(stream: BinaryData) -> Script:

    s = bytesio_from_binarydata(stream)
    # initialize the result list
    r: Script = []
    while True:
        # get one byte
        t = s.read(1)
        if not t:
            break
        # convert it to an integer
        i = t[0]
        if i == 0:
            # numeric value 0 (OP_0)
            # r.append(OP_CODE_NAMES[i])
            r.append(i)
        elif i == 79:
            # numeric value -1 (OP_1NEGATE)
            r.append(-1)
        elif 80 < i < 97:
            # numeric values 1-16 (OP_1-OP_16)
            # r.append(OP_CODE_NAMES[i])
            r.append(i - 80)
        elif i < 76:
            # 1-byte-data-length | data
            data = s.read(i)
            r.append(data.hex().upper())
        elif i == 76:
            # OP_PUSHDATA1 | 1-byte-data-length | data
            data_length = int.from_bytes(s.read(1), byteorder="little", signed=False)
            data = s.read(data_length)
            r.append(data.hex().upper())
        elif i == 77:
            # OP_PUSHDATA2 | 2-byte-data-length | data
            data_length = int.from_bytes(s.read(2), byteorder="little", signed=False)
            data = s.read(data_length)
            r.append(data.hex().upper())
        elif i == 78:
            # OP_PUSHDATA4 | 4-byte-data-length | data
            data_length = int.from_bytes(s.read(4), byteorder="little", signed=False)
            data = s.read(data_length)
            r.append(data.hex().upper())
        else:
            # OP_CODE
            r.append(OP_CODE_NAMES[i])

    return r



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"ScriptPubKey functions."

from typing import Callable, List, Optional, Tuple

from btclib.sec_point import point_from_octets

from . import var_bytes
from .alias import Octets, String
from .exceptions import BTClibValueError
from .hashes import hash160_from_key
from .script import serialize
from .to_pub_key import Key, pub_keyinfo_from_key
from .utils import bytes_from_octets, bytesio_from_binarydata, hash160, sha256

# 1. Hash/WitnessProgram from pub_key/script_pub_key

# hash160_from_key, hash160, and sha256
# are imported from hashes.py and utils.py


# 2. script_pub_key from Hash/WitnessProgram and vice versa


def _is_funct(assert_funct: Callable[[Octets], None], script_pub_key: Octets) -> bool:

    try:
        # if the assert function detects a problem, it must rise an Exception
        assert_funct(script_pub_key)
    # must always return a bool: all Exceptions are catched
    except Exception:  # pylint: disable=broad-except
        return False
    return True


def assert_p2pk(script_pub_key: Octets) -> None:
    script_pub_key = bytes_from_octets(script_pub_key, (35, 67))
    # p2pk [pub_key, OP_CHECKSIG]
    # 0x41{65-byte pub_key}AC
    # or
    # 0x21{33-byte pub_key}AC
    if script_pub_key[-1] != 0xAC:
        raise BTClibValueError("missing final OP_CHECKSIG")

    len_marker = script_pub_key[0]
    length = len(script_pub_key)
    if length == 35:
        if len_marker != 0x21:
            err_msg = f"invalid pub_key length marker: {len_marker}"
            err_msg += f" instead of {0x21}"
            raise BTClibValueError(err_msg)
    elif length == 67:
        if len_marker != 0x41:
            err_msg = f"invalid pub_key length marker: {len_marker}"
            err_msg += f" instead of {0x41}"
            raise BTClibValueError(err_msg)

    pub_key = script_pub_key[1:-1]
    point_from_octets(pub_key)


def is_p2pk(script_pub_key: Octets) -> bool:
    return _is_funct(assert_p2pk, script_pub_key)


def assert_p2pkh(script_pub_key: Octets) -> None:
    script_pub_key = bytes_from_octets(script_pub_key, 25)
    # p2pkh [OP_DUP, OP_HASH160, pub_key hash, OP_EQUALVERIFY, OP_CHECKSIG]
    # 0x76A914{20-byte pub_key_hash}88AC
    if script_pub_key[-2:] != b"\x88\xac":
        raise BTClibValueError("missing final OP_EQUALVERIFY, OP_CHECKSIG")
    if script_pub_key[:2] != b"\x76\xa9":
        raise BTClibValueError("missing leading OP_DUP, OP_HASH160")
    if script_pub_key[2] != 0x14:
        err_msg = f"invalid pub_key hash length marker: {script_pub_key[2]}"
        err_msg += f" instead of {0x14}"
        raise BTClibValueError(err_msg)


def is_p2pkh(script_pub_key: Octets) -> bool:
    return _is_funct(assert_p2pkh, script_pub_key)


def assert_p2sh(script_pub_key: Octets) -> None:
    script_pub_key = bytes_from_octets(script_pub_key, 23)
    # p2sh [OP_HASH160, redeem_script hash, OP_EQUAL]
    # 0xA914{20-byte redeem_script hash}87
    if script_pub_key[-1] != 0x87:
        raise BTClibValueError("missing final OP_EQUAL")
    if script_pub_key[0] != 0xA9:
        raise BTClibValueError("missing leading OP_HASH160")
    if script_pub_key[1] != 0x14:
        err_msg = f"invalid redeem script hash length marker: {script_pub_key[1]}"
        err_msg += f" instead of {0x14}"
        raise BTClibValueError(err_msg)


def is_p2sh(script_pub_key: Octets) -> bool:
    return _is_funct(assert_p2sh, script_pub_key)


def assert_p2ms(script_pub_key: Octets) -> None:
    script_pub_key = bytes_from_octets(script_pub_key)
    # p2ms [m, pub_keys, n, OP_CHECKMULTISIG]
    length = len(script_pub_key)
    if length < 37:
        raise BTClibValueError(f"invalid length {length}")
    if script_pub_key[-1] != 0xAE:
        raise BTClibValueError("missing final OP_CHECKMULTISIG")
    m = script_pub_key[0] - 80
    if not 0 < m < 17:
        raise BTClibValueError(f"invalid m in m-of-n: {m}")
    n = script_pub_key[-2] - 80
    if not m <= n < 17:
        raise BTClibValueError(f"invalid m-of-n: {m}-of-{n}")

    stream = bytesio_from_binarydata(script_pub_key[1:-2])
    for _ in range(n):
        pub_key = var_bytes.deserialize(stream)
        point_from_octets(pub_key)

    if stream.read(1):
        raise BTClibValueError("invalid extra data")


def is_p2ms(script_pub_key: Octets) -> bool:
    return _is_funct(assert_p2ms, script_pub_key)


def assert_nulldata(script_pub_key: Octets) -> None:
    script_pub_key = bytes_from_octets(script_pub_key)
    # nulldata [OP_RETURN, data]
    length = len(script_pub_key)
    if length == 0:
        raise BTClibValueError("null length")
    if script_pub_key[0] != 0x6A:
        raise BTClibValueError("missing leading OP_RETURN")

    if length == 78 or length >= 84:
        raise BTClibValueError(f"invalid length {length}")

    # OP_RETURN, data length, data up to 75 bytes max
    # 0x6A{1 byte data-length}{data (0-75 bytes)}
    if length < 78:
        if script_pub_key[1] != length - 2:
            raise BTClibValueError(f"invalid data length marker {script_pub_key[1]}")
    # OP_RETURN, OP_PUSHDATA1, data length, data min 76 bytes up to 80
    # 0x6A4C{1-byte data-length}{data (76-80 bytes)}
    elif script_pub_key[1] != 0x4C or script_pub_key[2] != length - 3:
        err_msg = f"invalid data length marker {script_pub_key[1:2].hex()}"
        raise BTClibValueError(err_msg)


def is_nulldata(script_pub_key: Octets) -> bool:
    return _is_funct(assert_nulldata, script_pub_key)


def assert_p2wpkh(script_pub_key: Octets) -> None:
    script_pub_key = bytes_from_octets(script_pub_key, 22)
    # p2wpkh [0, pub_key hash]
    # 0x0014{20-byte pub_key hash}
    if script_pub_key[0] != 0:
        err_msg = f"invalid witness version: {script_pub_key[0]}"
        err_msg += f" instead of {0}"
        raise BTClibValueError(err_msg)
    if script_pub_key[1] != 0x14:
        err_msg = f"invalid pub_key hash length marker: {script_pub_key[1]}"
        err_msg += f" instead of {0x14}"
        raise BTClibValueError(err_msg)


def is_p2wpkh(script_pub_key: Octets) -> bool:
    return _is_funct(assert_p2wpkh, script_pub_key)


def assert_p2wsh(script_pub_key: Octets) -> None:
    script_pub_key = bytes_from_octets(script_pub_key, 34)
    # p2wsh [0, redeem_script hash]
    # 0x0020{32-byte redeem_script hash}
    if script_pub_key[0] != 0:
        err_msg = f"invalid witness version: {script_pub_key[0]}"
        err_msg += f" instead of {0}"
        raise BTClibValueError(err_msg)
    if script_pub_key[1] != 0x20:
        err_msg = f"invalid redeem script hash length marker: {script_pub_key[1]}"
        err_msg += f" instead of {0x20}"
        raise BTClibValueError(err_msg)


def is_p2wsh(script_pub_key: Octets) -> bool:
    return _is_funct(assert_p2wsh, script_pub_key)


def script_pub_key_from_payload(script_type: str, payload: Octets) -> bytes:
    "Return the script_pub_key for the provided script_type and payload."

    script_type = script_type.lower()

    if script_type == "p2ms":
        script_pub_key = bytes_from_octets(payload) + b"\xae"
        if not is_p2ms(script_pub_key):
            raise BTClibValueError("invalid p2ms payload")
        return script_pub_key

    if script_type == "nulldata":
        payload = bytes_from_octets(payload)
        if len(payload) > 80:
            err_msg = f"invalid nulldata script length: {len(payload)} bytes "
            raise BTClibValueError(err_msg)
        return serialize(["OP_RETURN", payload])

    if script_type == "p2pk":
        payload = bytes_from_octets(payload, (33, 65))
        # TODO: check it is a valid pub_key
        return serialize([payload, "OP_CHECKSIG"])

    if script_type == "p2wsh":
        payload = bytes_from_octets(payload, 32)
        return serialize([0, payload])

    if script_type == "p2pkh":
        payload = bytes_from_octets(payload, 20)
        return serialize(
            [
                "OP_DUP",
                "OP_HASH160",
                payload,
                "OP_EQUALVERIFY",
                "OP_CHECKSIG",
            ]
        )

    if script_type == "p2sh":
        payload = bytes_from_octets(payload, 20)
        return serialize(["OP_HASH160", payload, "OP_EQUAL"])

    if script_type == "p2wpkh":
        payload = bytes_from_octets(payload, 20)
        return serialize([0, payload])

    raise BTClibValueError(f"unknown script_pub_key type: {script_type}")


def payload_from_script_pub_key(script_pub_key: Octets) -> Tuple[str, bytes]:
    "Return (script_pub_key type, payload) from the input script_pub_key."

    script_pub_key = bytes_from_octets(script_pub_key)

    if is_p2wpkh(script_pub_key):
        # p2wpkh [0, pub_key_hash]
        # 0x0014{20-byte pub_key_hash}
        return "p2wpkh", script_pub_key[2:]

    if is_p2wsh(script_pub_key):
        # p2wsh [0, script_hash]
        # 0x0020{32-byte script_hash}
        return "p2wsh", script_pub_key[2:]

    if is_p2pk(script_pub_key):
        # p2pk [pub_key, OP_CHECKSIG]
        # 0x41{65-byte pub_key}AC or 0x21{33-byte pub_key}AC
        return "p2pk", script_pub_key[1:-1]

    if is_p2ms(script_pub_key):
        # p2ms [m, pub_keys, n, OP_CHECKMULTISIG]
        return "p2ms", script_pub_key[:-1]

    if is_nulldata(script_pub_key):
        # nulldata [OP_RETURN, data]
        if len(script_pub_key) < 78:
            # OP_RETURN, data length, data up to 75 bytes max
            # 0x6A{1 byte data-length}{data (0-75 bytes)}
            return "nulldata", script_pub_key[2:]

        # OP_RETURN, OP_PUSHDATA1, data length, data min 76 bytes up to 80
        # 0x6A4C{1-byte data-length}{data (76-80 bytes)}
        return "nulldata", script_pub_key[3:]

    if is_p2pkh(script_pub_key):
        # p2pkh [OP_DUP, OP_HASH160, pub_key_hash, OP_EQUALVERIFY, OP_CHECKSIG]
        # 0x76A914{20-byte pub_key_hash}88AC
        length = len(script_pub_key)
        return "p2pkh", script_pub_key[3 : length - 2]

    if is_p2sh(script_pub_key):
        # p2sh [OP_HASH160, script_hash, OP_EQUAL]
        # 0xA914{20-byte script_hash}87
        length = len(script_pub_key)
        return "p2sh", script_pub_key[2 : length - 1]

    return "unknown", script_pub_key


# 1.+2. = 3. script_pub_key from key(s)/script


def p2pk(key: Key) -> bytes:
    "Return the p2pk script_pub_key of the provided key."

    payload, _ = pub_keyinfo_from_key(key)
    return script_pub_key_from_payload("p2pk", payload)


def p2ms(
    m: int, keys: List[Key], lexi_sort: bool = True, compressed: Optional[bool] = None
) -> bytes:
    """Return the m-of-n multi-sig script_pub_key of the provided keys.

    BIP67 endorses lexicographica key sorting
    according to compressed key representation.

    Note that sorting uncompressed keys (leading 0x04 byte) results
    in a different order than sorting the same keys in compressed
    (leading 0x02 or 0x03 bytes) representation.

    https://github.com/bitcoin/bips/blob/master/bip-0067.mediawiki
    """
    m += 80
    payload = m.to_bytes(1, byteorder="big", signed=False)
    pub_keys = [pub_keyinfo_from_key(k, compressed=compressed)[0] for k in keys]
    if lexi_sort:
        pub_keys = sorted(pub_keys)
    payload += b"".join([var_bytes.serialize(k) for k in pub_keys])
    n = len(keys) + 80
    payload += n.to_bytes(1, byteorder="big", signed=False)
    return script_pub_key_from_payload("p2ms", payload)


def nulldata(data: String) -> bytes:
    "Return the nulldata script_pub_key of the provided data."

    if isinstance(data, str):
        # do not strip spaces
        data = data.encode()

    return script_pub_key_from_payload("nulldata", data)


def p2pkh(key: Key, compressed: Optional[bool] = None) -> bytes:
    "Return the p2pkh script_pub_key of the provided key."

    pub_key_h160, _ = hash160_from_key(key, compressed=compressed)
    return script_pub_key_from_payload("p2pkh", pub_key_h160)


def p2sh(redeem_script: Octets) -> bytes:
    "Return the p2sh script_pub_key of the provided redeem script."

    script_h160 = hash160(redeem_script)
    return script_pub_key_from_payload("p2sh", script_h160)


def p2wpkh(key: Key) -> bytes:
    """Return the p2wpkh script_pub_key of the provided key.

    If the provided key is a public one, it must be compressed.
    """

    pub_key_h160, _ = hash160_from_key(key, compressed=True)
    return script_pub_key_from_payload("p2wpkh", pub_key_h160)


def p2wsh(redeem_script: Octets) -> bytes:
    "Return the p2wsh script_pub_key of the provided redeem script."

    script_h256 = sha256(redeem_script)
    return script_pub_key_from_payload("p2wsh", script_h256)



#!/usr/bin/env python3

# Copyright (C) 2020-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Transaction hashes to be signed and their hash types.

https://medium.com/@bitaps.com/exploring-bitcoin-signature-hash-types-15427766f0a9
https://raghavsood.com/blog/2018/06/10/bitcoin-signature-types-sighash
https://wiki.bitcoinsv.io/index.php/SIGHASH_flags
"""

from copy import deepcopy
from typing import List

from . import script, var_bytes
from .alias import Octets
from .exceptions import BTClibValueError
from .script_pub_key import (
    is_p2sh,
    is_p2wpkh,
    is_p2wsh,
    payload_from_script_pub_key,
    script_pub_key_from_payload,
)
from .tx import Tx
from .tx_out import TxOut
from .utils import bytes_from_octets, hash256

ALL = 1
NONE = 2
SINGLE = 3
ANYONECANPAY = 0b10000000

SIG_HASH_TYPES = [
    ALL,
    NONE,
    SINGLE,
    ANYONECANPAY & ALL,
    ANYONECANPAY & NONE,
    ANYONECANPAY & SINGLE,
]


def assert_valid_hash_type(hash_type: int) -> None:
    if hash_type not in SIG_HASH_TYPES:
        raise BTClibValueError(f"invalid sign_hash type: {hex(hash_type)}")


def _legacy_script(script_pub_key: Octets) -> List[bytes]:
    script_s: List[bytes] = []
    current_script: List[script.ScriptToken] = []
    for token in script.deserialize(script_pub_key)[::-1]:
        if token == "OP_CODESEPARATOR":  # nosec required for python < 3.8
            script_s.append(script.serialize(current_script[::-1]))
        else:
            current_script.append(token)
    script_s.append(script.serialize(current_script[::-1]))
    return script_s[::-1]


# FIXME: remove OP_CODESEPARATOR only if executed
def _witness_v0_script(script_pub_key: Octets) -> List[bytes]:
    script_type, payload = payload_from_script_pub_key(script_pub_key)

    if script_type == "p2wpkh":
        return [script_pub_key_from_payload("p2pkh", payload)]

    script_s: List[bytes] = []
    current_script: List[script.ScriptToken] = []
    for token in script.deserialize(script_pub_key)[::-1]:
        if token == "OP_CODESEPARATOR":  # nosec required for python < 3.8
            script_s.append(script.serialize(current_script[::-1]))
        current_script.append(token)
    script_s.append(script.serialize(current_script[::-1]))
    return script_s[::-1]


def legacy(script_: Octets, tx: Tx, vin_i: int, hash_type: int) -> bytes:
    script_ = bytes_from_octets(script_)

    new_tx = deepcopy(tx)
    for txin in new_tx.vin:
        txin.script_sig = b""
    # TODO: delete sig from script_ (even if non standard)
    new_tx.vin[vin_i].script_sig = script_
    if hash_type & 0x1F == NONE:
        new_tx.vout = []
        for i, txin in enumerate(new_tx.vin):
            if i != vin_i:
                txin.sequence = 0

    if hash_type & 0x1F == SINGLE:
        # sign_hash single bug
        if vin_i >= len(new_tx.vout):
            return (256 ** 31).to_bytes(32, byteorder="big", signed=False)
        new_tx.vout = new_tx.vout[: vin_i + 1]
        for txout in new_tx.vout[:-1]:
            txout.script_pub_key = b""
            txout.value = 256 ** 8 - 1
        for i, txin in enumerate(new_tx.vin):
            if i != vin_i:
                txin.sequence = 0

    if hash_type & 0x80:
        new_tx.vin = [new_tx.vin[vin_i]]

    preimage = new_tx.serialize(include_witness=False, check_validity=False)
    preimage += hash_type.to_bytes(4, byteorder="little", signed=False)

    return hash256(preimage)


# https://github.com/bitcoin/bitcoin/blob/4b30c41b4ebf2eb70d8a3cd99cf4d05d405eec81/test/functional/test_framework/script.py#L673
def segwit_v0(
    script_: Octets, tx: Tx, vin_i: int, hash_type: int, amount: int
) -> bytes:
    script_ = bytes_from_octets(script_)

    hash_prev_outs = b"\x00" * 32
    if not hash_type & ANYONECANPAY:
        hash_prev_outs = b"".join([vin.prev_out.serialize() for vin in tx.vin])
        hash_prev_outs = hash256(hash_prev_outs)

    hash_seqs = b"\x00" * 32
    if (
        not (hash_type & ANYONECANPAY)
        and (hash_type & 0x1F) != SINGLE
        and (hash_type & 0x1F) != NONE
    ):
        hash_seqs = b"".join(
            [
                vin.sequence.to_bytes(4, byteorder="little", signed=False)
                for vin in tx.vin
            ]
        )
        hash_seqs = hash256(hash_seqs)

    hash_outputs = b"\x00" * 32
    if hash_type & 0x1F not in (SINGLE, NONE):
        hash_outputs = b"".join([vout.serialize() for vout in tx.vout])
        hash_outputs = hash256(hash_outputs)
    elif (hash_type & 0x1F) == SINGLE and vin_i < len(tx.vout):
        hash_outputs = hash256(tx.vout[vin_i].serialize())

    preimage = tx.version.to_bytes(4, byteorder="little", signed=False)
    preimage += hash_prev_outs
    preimage += hash_seqs
    preimage += tx.vin[vin_i].prev_out.serialize()
    preimage += var_bytes.serialize(script_)
    preimage += amount.to_bytes(8, byteorder="little", signed=False)  # value
    preimage += tx.vin[vin_i].sequence.to_bytes(4, byteorder="little", signed=False)
    preimage += hash_outputs
    preimage += tx.lock_time.to_bytes(4, byteorder="little", signed=False)
    preimage += hash_type.to_bytes(4, byteorder="little", signed=False)

    return hash256(preimage)


def from_utxo(utxo: TxOut, tx: Tx, vin_i: int, hash_type: int) -> bytes:

    script_pub_key = utxo.script_pub_key

    # first off, handle all p2sh-wrapped scripts
    if is_p2sh(script_pub_key):
        script_pub_key = tx.vin[vin_i].script_sig

    if is_p2wpkh(script_pub_key):
        script_ = _witness_v0_script(script_pub_key)[0]
        return segwit_v0(script_, tx, vin_i, hash_type, utxo.value)

    if is_p2wsh(script_pub_key):
        # the real script is contained in the witness
        script_ = _witness_v0_script(tx.vin[vin_i].script_witness.stack[-1])[0]
        return segwit_v0(script_, tx, vin_i, hash_type, utxo.value)

    script_ = _legacy_script(script_pub_key)[0]
    return legacy(script_, tx, vin_i, hash_type)



#!/usr/bin/env python3

# Copyright (C) 2020-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Bitcoin Transaction.

https://en.bitcoin.it/wiki/Transaction
https://learnmeabitcoin.com/guide/coinbase-transaction
https://bitcoin.stackexchange.com/questions/20721/what-is-the-format-of-the-coinbase-transaction

For TxIn.sequence and TX.lock_time see:
https://developer.bitcoin.org/devguide/transactions.html
https://medium.com/summa-technology/bitcoins-time-locks-27e0c362d7a1
https://bitcoin.stackexchange.com/questions/40764/is-my-understanding-of-locktime-correct
https://en.bitcoin.it/wiki/Timelock

"""

from dataclasses import InitVar, dataclass, field
from io import SEEK_CUR
from math import ceil
from typing import Dict, List, Type, TypeVar

from dataclasses_json import DataClassJsonMixin, config
from dataclasses_json.core import Json

from . import var_int
from .alias import BinaryData
from .exceptions import BTClibValueError
from .tx_in import _TX_IN_COMPARES_WITNESS, TxIn
from .tx_out import TxOut
from .utils import bytesio_from_binarydata, hash256
from .witness import Witness

_SEGWIT_MARKER = b"\x00\x01"

_Tx = TypeVar("_Tx", bound="Tx")


@dataclass
class Tx(DataClassJsonMixin):
    # private data members are used only for to_dict
    # use the corresponding public properties instead
    _tx_id: bytes = field(
        default=b"",
        init=False,
        repr=False,
        compare=False,
        metadata=config(
            encoder=lambda v: v.hex(), decoder=bytes.fromhex, field_name="txid"
        ),
    )
    _hash: bytes = field(
        default=b"",
        init=False,
        repr=False,
        compare=False,
        metadata=config(
            encoder=lambda v: v.hex(), decoder=bytes.fromhex, field_name="hash"
        ),
    )
    # 4 bytes, unsigned little endian
    version: int = 1
    _size: int = field(
        default=-1,
        init=False,
        repr=False,
        compare=False,
        metadata=config(field_name="size"),
    )
    _vsize: int = field(
        default=-1,
        init=False,
        repr=False,
        compare=False,
        metadata=config(field_name="vsize"),
    )
    _weight: int = field(
        default=-1,
        init=False,
        repr=False,
        compare=False,
        metadata=config(field_name="weight"),
    )
    # 0	Not locked
    #  < 500000000	Block number at which this transaction is unlocked
    # >= 500000000	UNIX timestamp at which this transaction is unlocked
    # If all TxIns have final (0xffffffff) sequence numbers then lock_time is irrelevant.
    # Otherwise, the transaction may not be added to a block until after lock_time.
    # Set to the current block to prevent fee sniping.
    lock_time: int = field(
        default=0,
        metadata=config(field_name="locktime"),
    )
    vin: List[TxIn] = field(default_factory=list)
    vout: List[TxOut] = field(default_factory=list)
    # TODO: add fee when a tx fetcher will be available
    check_validity: InitVar[bool] = True

    def __post_init__(self, check_validity: bool) -> None:
        if check_validity:
            self.assert_valid()

    def __eq__(self, other) -> bool:
        if not isinstance(other, Tx):
            return NotImplemented  # pragma: no cover

        if not _TX_IN_COMPARES_WITNESS and self.vwitness != other.vwitness:
            return False  # pragma: no cover

        return (self.version, self.lock_time, self.vin, self.vout) == (
            other.version,
            other.lock_time,
            other.vin,
            other.vout,
        )

    def _set_properties(self) -> None:
        self._tx_id = self.tx_id
        self._hash = self.hash
        self._size = self.size
        self._vsize = self.vsize
        self._weight = self.weight

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        self._set_properties()
        return super().to_dict(encode_json)

    @property
    def nVersion(self) -> int:  # pylint: disable=invalid-name
        "Return the nVersion int for compatibility with CTransaction."
        return self.version

    @property
    def nLockTime(self) -> int:  # pylint: disable=invalid-name
        "Return the nLockTime int for compatibility with CTransaction."
        return self.lock_time

    @property
    def tx_id(self) -> bytes:
        serialized_ = self.serialize(include_witness=False, check_validity=False)
        hash256_ = hash256(serialized_)
        return hash256_[::-1]

    @property
    def hash(self) -> bytes:
        serialized_ = self.serialize(include_witness=True, check_validity=False)
        hash256_ = hash256(serialized_)
        return hash256_[::-1]

    @property
    def size(self) -> int:
        return len(self.serialize(include_witness=True, check_validity=False))

    @property
    def vsize(self) -> int:
        return ceil(self.weight / 4)

    @property
    def weight(self) -> int:
        no_wit = len(self.serialize(include_witness=False, check_validity=False)) * 3
        wit = len(self.serialize(include_witness=True, check_validity=False))
        return no_wit + wit

    @property
    def vwitness(self) -> List[Witness]:
        return [tx_in.script_witness for tx_in in self.vin]

    def is_segwit(self) -> bool:
        # refer to tx_in, not tx_out
        return any(tx_in.is_segwit() for tx_in in self.vin)

    def is_coinbase(self) -> bool:
        return len(self.vin) == 1 and self.vin[0].is_coinbase()

    def assert_valid(self) -> None:

        # must be a 4-bytes int
        if not 0 < self.version <= 0xFFFFFFFF:
            raise BTClibValueError(f"invalid version: {self.version}")

        # must be a 4-bytes int
        if not 0 <= self.lock_time <= 0xFFFFFFFF:
            raise BTClibValueError(f"invalid lock time: {self.lock_time}")

        for tx_in in self.vin:
            tx_in.assert_valid()

        for tx_out in self.vout:
            tx_out.assert_valid()

        self._set_properties()

    def serialize(self, include_witness: bool, check_validity: bool = True) -> bytes:

        if check_validity:
            self.assert_valid()

        segwit = include_witness and self.is_segwit()

        out = self.version.to_bytes(4, byteorder="little", signed=False)
        out += _SEGWIT_MARKER if segwit else b""
        out += var_int.serialize(len(self.vin))
        out += b"".join(tx_in.serialize(check_validity) for tx_in in self.vin)
        out += var_int.serialize(len(self.vout))
        out += b"".join(tx_out.serialize(check_validity) for tx_out in self.vout)
        if segwit:
            out += b"".join(
                tx_in.script_witness.serialize(check_validity) for tx_in in self.vin
            )
        out += self.lock_time.to_bytes(4, byteorder="little", signed=False)

        return out

    @classmethod
    def deserialize(
        cls: Type[_Tx], data: BinaryData, check_validity: bool = True
    ) -> _Tx:
        "Return a Tx by parsing binary data."

        stream = bytesio_from_binarydata(data)

        version = int.from_bytes(stream.read(4), byteorder="little", signed=False)

        segwit = stream.read(2) == _SEGWIT_MARKER
        if not segwit:
            # Change stream position: seek to byte offset relative to position
            stream.seek(-2, SEEK_CUR)  # current position

        n = var_int.deserialize(stream)
        vin = [TxIn.deserialize(stream) for _ in range(n)]

        n = var_int.deserialize(stream)
        vout = [TxOut.deserialize(stream) for _ in range(n)]

        if segwit:
            for tx_in in vin:
                tx_in.script_witness = Witness.deserialize(stream, check_validity)

        lock_time = int.from_bytes(stream.read(4), byteorder="little", signed=False)

        return cls(version, lock_time, vin, vout, check_validity)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.bip32_path` module."

import json
from os import path

import pytest

from btclib import bip32, hashes
from btclib.bip32_path import (
    _HARDENING,
    BIP32KeyOrigin,
    BIP32KeyPath,
    BIP32KeyPaths,
    _assert_valid_hd_key_paths,
    _decode_hd_key_paths,
    _indexes_from_bip32_path_str,
    _int_from_index_str,
    _serialize_hd_key_paths,
    _str_from_index_int,
    bytes_from_bip32_path,
    indexes_from_bip32_path,
    str_from_bip32_path,
)
from btclib.exceptions import BTClibValueError


def test_indexes_from_bip32_path_str() -> None:

    test_reg_str_vectors = [
        # account 0, external branch, address_index 463
        ("m/0" + _HARDENING + "/0/463", [0x80000000, 0, 463]),
        # account 0, internal branch, address_index 267
        ("m/0" + _HARDENING + "/1/267", [0x80000000, 1, 267]),
    ]

    for bip32_path_str, bip32_path_ints in test_reg_str_vectors:
        # recover ints from str
        assert bip32_path_ints == _indexes_from_bip32_path_str(bip32_path_str)
        assert bip32_path_ints == indexes_from_bip32_path(bip32_path_str)
        # recover ints from ints
        assert bip32_path_ints == indexes_from_bip32_path(bip32_path_ints)
        # recover str from str
        assert bip32_path_str == str_from_bip32_path(bip32_path_str)
        # recover str from ints
        assert bip32_path_str == str_from_bip32_path(bip32_path_ints)
        # ensure bytes from ints == bytes from str
        bip32_path_bytes = bytes_from_bip32_path(bip32_path_ints)
        assert bip32_path_bytes == bytes_from_bip32_path(bip32_path_str)
        # recover ints from bytes
        assert bip32_path_ints == indexes_from_bip32_path(bip32_path_bytes)
        # recover str from bytes
        assert bip32_path_str == str_from_bip32_path(bip32_path_bytes)

    test_irregular_str_vectors = [
        # account 0, external branch, address_index 463
        ("m / 0 h / 0 / 463", [0x80000000, 0, 463]),
        ("m / 0 H / 0 / 463", [0x80000000, 0, 463]),
        ("m // 0' / 0 / 463", [0x80000000, 0, 463]),
        # account 0, internal branch, address_index 267
        ("m / 0 h / 1 / 267", [0x80000000, 1, 267]),
        ("m / 0 H / 1 / 267", [0x80000000, 1, 267]),
        ("m // 0' / 1 / 267", [0x80000000, 1, 267]),
    ]

    for bip32_path_str, bip32_path_ints in test_irregular_str_vectors:
        # recover ints from str
        assert bip32_path_ints == _indexes_from_bip32_path_str(bip32_path_str)
        assert bip32_path_ints == indexes_from_bip32_path(bip32_path_str)
        # recover ints from ints
        assert bip32_path_ints == indexes_from_bip32_path(bip32_path_ints)
        # irregular str != normalized str
        assert bip32_path_str != str_from_bip32_path(bip32_path_str)
        # irregular str != normalized str from ints
        assert bip32_path_str != str_from_bip32_path(bip32_path_ints)
        # ensure bytes from ints == bytes from str
        bip32_path_bytes = bytes_from_bip32_path(bip32_path_ints)
        assert bip32_path_bytes == bytes_from_bip32_path(bip32_path_str)
        # recover ints from bytes
        assert bip32_path_ints == indexes_from_bip32_path(bip32_path_bytes)
        # irregular str != normalized str from bytes
        assert bip32_path_str != str_from_bip32_path(bip32_path_bytes)

    with pytest.raises(BTClibValueError, match="invalid index: "):
        _indexes_from_bip32_path_str("m/1/2/-3h/4")

    with pytest.raises(BTClibValueError, match="invalid index: "):
        _indexes_from_bip32_path_str("m/1/2/-3/4")

    i = 0x80000000

    with pytest.raises(BTClibValueError, match="invalid index: "):
        _indexes_from_bip32_path_str("m/1/2/" + str(i) + "/4")

    with pytest.raises(BTClibValueError, match="invalid index: "):
        _indexes_from_bip32_path_str("m/1/2/" + str(i) + "h/4")


def test_index_int_to_from_str() -> None:

    for i in (0, 1, 0x80000000 - 1, 0x80000000, 0xFFFFFFFF):
        assert i == _int_from_index_str(_str_from_index_int(i))

    for i in (-1, 0xFFFFFFFF + 1):
        with pytest.raises(BTClibValueError):
            _str_from_index_int(i)

    for s in ("-1", "-1h", str(0x80000000) + "h", str(0xFFFFFFFF + 1)):
        with pytest.raises(BTClibValueError):
            _int_from_index_str(s)

    with pytest.raises(BTClibValueError, match="invalid hardening symbol: "):
        _str_from_index_int(0x80000000, "hardened")


def test_bip32_key_origin() -> None:

    invalid_key_origin = BIP32KeyOrigin(
        bytes.fromhex("badbad"), [0], check_validity=False
    )
    with pytest.raises(BTClibValueError, match="invalid master fingerprint length: "):
        invalid_key_origin.assert_valid()

    invalid_key_origin = BIP32KeyOrigin(
        bytes.fromhex("deadbeef"), [0] * 256, check_validity=False
    )
    with pytest.raises(BTClibValueError, match="invalid der_path size: "):
        invalid_key_origin.assert_valid()

    invalid_key_origin = BIP32KeyOrigin(
        bytes.fromhex("deadbeef"), [0xFFFFFFFF + 1], check_validity=False
    )
    with pytest.raises(BTClibValueError, match="invalid der_path element"):
        invalid_key_origin.assert_valid()

    fingerprint = "deadbeef"
    description = fingerprint
    key_origin = BIP32KeyOrigin.from_description(description)
    key_origin2 = BIP32KeyOrigin.from_description(description + "/")
    assert key_origin == key_origin2
    assert key_origin.description == description
    assert key_origin.fingerprint == bytes.fromhex(fingerprint)
    assert key_origin.der_path == []
    assert BIP32KeyOrigin.deserialize(key_origin.serialize()) == key_origin

    description = (  # use the hardening convention of the normalized der_path
        fingerprint
        + "/44"
        + _HARDENING
        + "/0"
        + _HARDENING
        + "/1"
        + _HARDENING
        + "/0/10"
    )
    key_origin = BIP32KeyOrigin.from_description(description)
    key_origin2 = BIP32KeyOrigin.from_description("deadbeef//44h/0'/1H/0/10/")
    assert key_origin == key_origin2
    assert key_origin.description == description
    assert key_origin.fingerprint == bytes.fromhex(fingerprint)
    assert key_origin.der_path == [
        44 + 0x80000000,
        0 + 0x80000000,
        1 + 0x80000000,
        0,
        10,
    ]
    assert BIP32KeyOrigin.deserialize(key_origin.serialize()) == key_origin
    assert len(key_origin) == 5


def test_dataclasses_json_dict_key_origin() -> None:

    key_origin = BIP32KeyOrigin.from_description("deadbeef//44h/0'/1H/0/10/")

    # BIP32KeyOrigin dataclass
    assert isinstance(key_origin, BIP32KeyOrigin)
    key_origin.assert_valid()

    # BIP32KeyOrigin dataclass to dict
    key_origin_dict = key_origin.to_dict()
    assert isinstance(key_origin_dict, dict)
    assert key_origin_dict["master_fingerprint"]
    assert key_origin_dict["path"]

    # BIP32KeyOrigin dataclass dict to file
    datadir = path.join(path.dirname(__file__), "generated_files")
    filename = path.join(datadir, "key_origin.json")
    with open(filename, "w") as file_:
        json.dump(key_origin_dict, file_, indent=4)

    # BIP32KeyOrigin dataclass dict from file
    with open(filename, "r") as file_:
        key_origin_dict2 = json.load(file_)
    assert isinstance(key_origin_dict2, dict)
    assert key_origin_dict["master_fingerprint"]
    assert key_origin_dict["path"]

    assert key_origin_dict == key_origin_dict2

    # BIP32KeyOrigin dataclass from dict
    key_origin2 = BIP32KeyOrigin.from_dict(key_origin_dict)
    assert isinstance(key_origin2, BIP32KeyOrigin)
    key_origin2.assert_valid()

    assert key_origin == key_origin2


def test_bip32_key_path() -> None:

    pub_key = bytes.fromhex(
        "02 50863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352"
    )
    invalid_key_origin = BIP32KeyOrigin(
        bytes.fromhex("badbad"), [0], check_validity=False
    )
    invalid_key_path = BIP32KeyPath(pub_key, invalid_key_origin, check_validity=False)
    with pytest.raises(BTClibValueError, match="invalid master fingerprint length: "):
        invalid_key_path.assert_valid()

    fingerprint = "deadbeef"
    der_path = "/44h/0h/1h/0/10"
    description = fingerprint + der_path
    key_origin = BIP32KeyOrigin.from_description(description)
    assert len(key_origin) == 5

    key_path = BIP32KeyPath(pub_key, key_origin)
    assert key_path == BIP32KeyPath.deserialize(key_path.serialize())
    assert len(key_path) == len(key_origin)


def test_dataclasses_json_dict_key_path() -> None:

    fingerprint = "deadbeef"
    der_path = "/44'/0'/1'/0/10"
    description = fingerprint + der_path
    key_origin = BIP32KeyOrigin.from_description(description)
    pub_key = bytes.fromhex(
        "02 50863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352"
    )
    key_path = BIP32KeyPath(pub_key, key_origin)

    # BIP32KeyPath dataclass
    assert isinstance(key_path, BIP32KeyPath)

    # BIP32KeyPath dataclass to dict
    key_path_dict = key_path.to_dict()
    assert isinstance(key_path_dict, dict)
    assert key_path_dict["pub_key"]
    assert key_path_dict["key_origin"]

    # BIP32KeyPath dataclass dict to file
    datadir = path.join(path.dirname(__file__), "generated_files")
    filename = path.join(datadir, "key_path.json")
    with open(filename, "w") as file_:
        json.dump(key_path_dict, file_, indent=4)

    # BIP32KeyPath dataclass dict from file
    with open(filename, "r") as file_:
        key_path_dict2 = json.load(file_)
    assert isinstance(key_path_dict2, dict)
    assert key_path_dict2["pub_key"]
    assert key_path_dict2["key_origin"]

    assert key_path_dict == key_path_dict2

    # BIP32KeyPath dataclass from dict
    key_path2 = BIP32KeyPath.from_dict(key_path_dict)
    assert isinstance(key_path2, BIP32KeyPath)

    assert key_path == key_path2


_MARKER = b"\x11"


def test_bip32_derivs() -> None:
    # the basic type dict representation
    encoded_data = [
        {
            "pub_key": "029583bf39ae0a609747ad199addd634fa6108559d6c5cd39b4c2183f1ab96e07f",
            "key_origin": {
                "master_fingerprint": "d90c6a4f",
                "path": "m/0" + _HARDENING + "/0/0",
            },
        },
        {
            "pub_key": "02dab61ff49a14db6a7d02b0cd1fbb78fc4b18312b5b4e54dae4dba2fbfef536d7",
            "key_origin": {
                "master_fingerprint": "d90c6a4f",
                "path": "m/0" + _HARDENING + "/0/1",
            },
        },
    ]
    data = _decode_hd_key_paths(encoded_data)
    _assert_valid_hd_key_paths(data)
    # assert encoded_data == _encode_hd_key_paths(data)

    _serialize_hd_key_paths(_MARKER, data)
    # TODO: check deserialization


def test_bip32_key_paths() -> None:

    master_xprv = "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
    fingerprint = hashes.fingerprint(master_xprv)

    der_path0 = "m/44h/0h/1h/0/0"
    der_path1 = "m/44h/0h/1h/0/1"
    der_path2 = "m/44h/0h/1h/0/2"

    pub_key0 = bip32._xpub_from_xprv(bip32._derive(master_xprv, der_path0))
    pub_key1 = bip32._xpub_from_xprv(bip32._derive(master_xprv, der_path1))
    pub_key2 = bip32._xpub_from_xprv(bip32._derive(master_xprv, der_path2))

    hd_key_paths = {
        # FIXME: allow BIP32KeyOrigin with der_path string
        pub_key0.key: BIP32KeyOrigin(fingerprint, indexes_from_bip32_path(der_path0)),
        pub_key1.key: BIP32KeyOrigin(fingerprint, indexes_from_bip32_path(der_path1)),
        pub_key2.key: BIP32KeyOrigin(fingerprint, indexes_from_bip32_path(der_path2)),
    }

    key_paths = BIP32KeyPaths(hd_key_paths)
    assert len(key_paths) == len(hd_key_paths)

    key_paths_bin = key_paths.serialize(b"\x00")
    key_paths_deser = BIP32KeyPaths.deserialize(key_paths_bin, b"\x00")
    assert len(key_paths_deser) == len(hd_key_paths)
    assert key_paths == key_paths_deser

    with pytest.raises(BTClibValueError, match="invalid type marker"):
        key_paths.serialize(b"\x00\x00")
    with pytest.raises(BTClibValueError, match="invalid type marker"):
        BIP32KeyPaths.deserialize(key_paths_bin, b"\x00\x00")


def test_dataclasses_json_dict_key_paths() -> None:

    master_xprv = "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
    fingerprint = hashes.fingerprint(master_xprv)

    der_path0 = "m/44h/0h/1h/0/0"
    der_path1 = "m/44h/0h/1h/0/1"
    der_path2 = "m/44h/0h/1h/0/2"

    pub_key0 = bip32._xpub_from_xprv(bip32._derive(master_xprv, der_path0))
    pub_key1 = bip32._xpub_from_xprv(bip32._derive(master_xprv, der_path1))
    pub_key2 = bip32._xpub_from_xprv(bip32._derive(master_xprv, der_path2))

    hd_key_paths = {
        # FIXME: allow BIP32KeyOrigin with der_path string
        pub_key0.key: BIP32KeyOrigin(fingerprint, indexes_from_bip32_path(der_path0)),
        pub_key1.key: BIP32KeyOrigin(fingerprint, indexes_from_bip32_path(der_path1)),
        pub_key2.key: BIP32KeyOrigin(fingerprint, indexes_from_bip32_path(der_path2)),
    }

    key_paths = BIP32KeyPaths(hd_key_paths)

    # BIP32KeyPaths dataclass
    assert isinstance(key_paths, BIP32KeyPaths)

    # BIP32KeyPaths dataclass to dict
    key_path_dicts = key_paths.to_dict()
    assert isinstance(key_path_dicts, dict)

    # BIP32KeyPaths dataclass dict to file
    datadir = path.join(path.dirname(__file__), "generated_files")
    filename = path.join(datadir, "key_paths.json")
    with open(filename, "w") as file_:
        json.dump(key_path_dicts, file_, indent=4)

    # BIP32KeyPaths dataclass dict from file
    with open(filename, "r") as file_:
        key_path_dict2 = json.load(file_)
    assert isinstance(key_path_dict2, dict)

    assert key_path_dicts == key_path_dict2

    # BIP32KeyPaths dataclass from dict
    key_path2 = BIP32KeyPaths.from_dict(key_path_dicts)
    assert isinstance(key_path2, BIP32KeyPaths)

    assert key_paths == key_path2



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.amount` module."

from btclib import amount


def test_amount_conversion() -> None:
    v1 = 1.1
    v2 = 2.2
    vtot = v1 + v2
    # _NOT_ equal !!
    assert vtot != 3.3
    s_1 = amount.sat_from_float(v1)
    s_2 = amount.sat_from_float(v2)
    stot = s_1 + s_2
    assert stot == 330000000
    vtot = amount.float_from_sat(stot)
    # equal !!
    assert vtot == 3.3



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.sign_to_contract` module."

import secrets
from hashlib import sha256

from btclib import dsa, ssa
from btclib.curve import secp256k1
from btclib.sign_to_contract import (
    ecdsa_commit_sign,
    ecssa_commit_sign,
    verify_commit,
)

ec = secp256k1


def test_sign_to_contract_dsa() -> None:
    m = sha256(b"to be signed").digest()
    c = sha256(b"to be committed").digest()

    prv_key, pub_key = dsa.gen_keys()
    dsa_sig, dsa_receipt = ecdsa_commit_sign(c, m, prv_key)
    dsa._assert_as_valid(m, pub_key, dsa_sig, sha256)
    assert verify_commit(c, dsa_receipt)

    k = 1 + secrets.randbelow(ec.n - 1)
    dsa_sig, dsa_receipt = ecdsa_commit_sign(c, m, prv_key, k)
    dsa._assert_as_valid(m, pub_key, dsa_sig, sha256)
    assert verify_commit(c, dsa_receipt)


def test_sign_to_contract_ssa() -> None:
    m = sha256(b"to be signed").digest()
    c = sha256(b"to be committed").digest()

    prv_key, pub = ssa.gen_keys()
    ssa_sig, ssa_receipt = ecssa_commit_sign(c, m, prv_key)
    ssa._assert_as_valid(m, pub, ssa_sig, sha256)
    assert verify_commit(c, ssa_receipt)

    k = 1 + secrets.randbelow(ec.n - 1)
    ssa_sig, ssa_receipt = ecssa_commit_sign(c, m, prv_key, k)
    ssa._assert_as_valid(m, pub, ssa_sig, sha256)
    assert verify_commit(c, ssa_receipt)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.hashes` module."

from btclib.bip32 import BIP32KeyData, derive, rootxprv_from_seed
from btclib.hashes import fingerprint, tagged_hash


def test_fingerprint() -> None:

    seed = "bfc4cbaad0ff131aa97fa30a48d09ae7df914bcc083af1e07793cd0a7c61a03f65d622848209ad3366a419f4718a80ec9037df107d8d12c19b83202de00a40ad"
    xprv = rootxprv_from_seed(seed)
    pf = fingerprint(xprv)  # xprv is automatically converted to xpub
    child_key = derive(xprv, 0x80000000)
    pf2 = BIP32KeyData.b58decode(child_key).parent_fingerprint
    assert pf == pf2


def test_tagged_hash() -> None:

    s = "deadbeef"
    b = s.encode()
    assert tagged_hash("btclib", s) == tagged_hash("btclib", b)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.script_pub_key` module."

import json
from os import path
from typing import List

import pytest

from btclib import base58_address, bech32_address, script, var_bytes
from btclib.base58_address import (
    base58_address_from_h160,
    base58_address_from_witness,
)
from btclib.bech32_address import bech32_address_from_witness
from btclib.exceptions import BTClibValueError
from btclib.network import NETWORKS
from btclib.script import Script
from btclib.script_pub_key import (
    assert_p2ms,
    assert_p2pk,
    assert_p2pkh,
    assert_p2sh,
    assert_p2wpkh,
    assert_p2wsh,
    is_nulldata,
    is_p2ms,
    nulldata,
    p2ms,
    p2pk,
    p2pkh,
    p2sh,
    p2wpkh,
    p2wsh,
    payload_from_script_pub_key,
    script_pub_key_from_payload,
)
from btclib.script_pub_key_address import (
    address_from_script_pub_key,
    script_pub_key_from_address,
)
from btclib.to_pub_key import Key
from btclib.utils import hash160, sha256


def test_nulldata() -> None:

    OP_RETURN = b"\x6a"  # pylint: disable=invalid-name

    # self-consistency
    string = "time-stamped data"
    payload = string.encode()
    script_pub_key = script.serialize(["OP_RETURN", payload])
    assert script_pub_key == nulldata(string)

    # to the script_pub_key in two steps (through payload)
    script_type = "nulldata"
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)

    # back from the script_pub_key to the payload
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)

    # data -> payload in this case is invertible (no hash functions)
    assert payload.decode("ascii") == string

    assert address_from_script_pub_key(script_pub_key) == ""

    # documented test cases: https://learnmeabitcoin.com/guide/nulldata
    string = "hello world"
    payload = string.encode()
    assert payload.hex() == "68656c6c6f20776f726c64"  # pylint: disable=no-member
    script_pub_key = OP_RETURN + var_bytes.serialize(payload)
    assert script_pub_key == nulldata(string)
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)

    # documented test cases: https://learnmeabitcoin.com/guide/nulldata
    string = "charley loves heidi"
    payload = string.encode()
    assert (
        payload.hex()  # pylint: disable=no-member
        == "636861726c6579206c6f766573206865696469"
    )
    script_pub_key = OP_RETURN + var_bytes.serialize(payload)
    assert script_pub_key == nulldata(string)
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)

    # documented test cases: https://learnmeabitcoin.com/guide/nulldata
    string = "家族も友達もみんなが笑顔の毎日がほしい"
    payload = string.encode()
    assert (
        payload.hex()  # pylint: disable=no-member
        == "e5aeb6e6978fe38282e58f8be98194e38282e381bfe38293e381aae3818ce7ac91e9a194e381aee6af8ee697a5e3818ce381bbe38197e38184"
    )
    script_pub_key = OP_RETURN + var_bytes.serialize(payload)
    assert script_pub_key == nulldata(string)
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)


def test_nulldata2() -> None:

    script_type = "nulldata"

    # max length case
    byte = b"\x00"
    for length in (0, 1, 16, 17, 74, 75, 76, 77, 78, 79, 80):
        payload = byte * length
        script_pub_key = script.serialize(["OP_RETURN", payload])
        assert script_pub_key == script_pub_key_from_payload(script_type, payload)

        # back from the script_pub_key to the payload
        assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)


def test_nulldata3() -> None:

    err_msg = "invalid nulldata script length: "
    with pytest.raises(BTClibValueError, match=err_msg):
        payload = "00" * 81
        script_pub_key_from_payload("nulldata", payload)

    # wrong data length: 32 in 35-bytes nulldata script;
    # it should have been 33
    script_pub_key = script.serialize(["OP_RETURN", b"\x00" * 33])
    script_pub_key = script_pub_key[:1] + b"\x20" + script_pub_key[2:]
    assert not is_nulldata(script_pub_key)

    # wrong data length: 32 in 83-bytes nulldata script;
    # it should have been 80
    script_pub_key = script.serialize(["OP_RETURN", b"\x00" * 80])
    script_pub_key = script_pub_key[:2] + b"\x20" + script_pub_key[3:]
    assert not is_nulldata(script_pub_key)

    # missing OP_PUSHDATA1 (0x4c) in 83-bytes nulldata script,
    # got 0x20 instead
    script_pub_key = script.serialize(["OP_RETURN", b"\x00" * 80])
    script_pub_key = script_pub_key[:1] + b"\x20" + script_pub_key[2:]
    assert not is_nulldata(script_pub_key)

    assert len(script.serialize(["OP_RETURN", b"\x00" * 75])) == 77
    assert len(script.serialize(["OP_RETURN", b"\x00" * 76])) == 79
    script_pub_key = script.serialize(["OP_RETURN", b"\x00" * 76])[:-1]
    assert not is_nulldata(script_pub_key)


def test_nulldata4() -> None:

    script_: Script = ["OP_RETURN", "OP_RETURN", 3, 1, "OP_VERIF", 0, 3]
    # FIXME: serialization is not 0x6A{1 byte data-length}{data 6 bytes)}
    script_pub_key = script.serialize(script_)
    assert len(script_pub_key) == 7
    assert script.deserialize(script_pub_key) == script_
    script_type, _ = payload_from_script_pub_key(script_pub_key)
    # FIXME: it should be "nulldata"
    assert script_type == "unknown"
    # assert is_nulldata(script_pub_key)


def test_p2pk() -> None:

    # self-consistency
    pub_key = "02 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
    script_pub_key = script.serialize([pub_key, "OP_CHECKSIG"])
    assert_p2pk(script_pub_key)
    assert script_pub_key == p2pk(pub_key)

    script_type = "p2pk"
    assert script_pub_key == script_pub_key_from_payload(script_type, pub_key)

    # back from the script_pub_key to the payload
    assert (script_type, bytes.fromhex(pub_key)) == payload_from_script_pub_key(
        script_pub_key
    )

    assert address_from_script_pub_key(script_pub_key) == ""

    err_msg = "invalid pub_key length marker: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2pk(b"\x31" + script_pub_key[1:])

    # documented test case: https://learnmeabitcoin.com/guide/p2pk
    pub_key = (
        "04"
        "ae1a62fe09c5f51b13905f07f06b99a2f7159b2225f374cd378d71302fa28414"
        "e7aab37397f554a7df5f142c21c1b7303b8a0626f1baded5c72a704f7e6cd84c"
    )
    script_pub_key = bytes.fromhex("41" + pub_key + "ac")
    assert_p2pk(script_pub_key)
    assert script_pub_key == p2pk(pub_key)

    err_msg = "missing final OP_CHECKSIG"
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2pk(script_pub_key[:-1] + b"\x00")

    err_msg = "invalid pub_key length marker: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2pk(b"\x31" + script_pub_key[1:])

    # invalid size: 34 bytes instead of (33, 65)
    pub_key = "03 ae1a62fe09c5f51b13905f07f06b99a2f7159b2225f374cd378d71302fa28414 14"
    err_msg = "not a private or public key: "
    with pytest.raises(BTClibValueError, match=err_msg):
        p2pk(pub_key)


def test_p2pkh() -> None:

    # self-consistency
    pub_key = (
        "04 "
        "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
        "f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
    )
    payload = hash160(pub_key)
    script_pub_key = script.serialize(
        ["OP_DUP", "OP_HASH160", payload, "OP_EQUALVERIFY", "OP_CHECKSIG"]
    )
    assert_p2pkh(script_pub_key)
    assert script_pub_key == p2pkh(pub_key)

    # to the script_pub_key in two steps (through payload)
    script_type = "p2pkh"
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)

    # back from the script_pub_key to the payload
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)

    # base58 address
    network = "mainnet"
    address = base58_address.p2pkh(pub_key, network)
    assert address == address_from_script_pub_key(script_pub_key, network)
    prefix = NETWORKS[network].p2pkh
    assert address == base58_address_from_h160(prefix, payload, network)

    # back from the address to the script_pub_key
    assert (script_pub_key, network) == script_pub_key_from_address(address)

    # documented test case: https://learnmeabitcoin.com/guide/p2pkh
    payload = bytes.fromhex("12ab8dc588ca9d5787dde7eb29569da63c3a238c")
    script_pub_key = bytes.fromhex("76a914") + payload + bytes.fromhex("88ac")
    assert_p2pkh(script_pub_key)
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)
    address = "12higDjoCCNXSA95xZMWUdPvXNmkAduhWv"
    assert address == address_from_script_pub_key(script_pub_key, network)
    assert (script_pub_key, network) == script_pub_key_from_address(address)

    err_msg = "missing final OP_EQUALVERIFY, OP_CHECKSIG"
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2pkh(script_pub_key[:-2] + b"\x40\x40")

    err_msg = "missing leading OP_DUP, OP_HASH160"
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2pkh(b"\x40\x40" + script_pub_key[2:])

    err_msg = "invalid pub_key hash length marker: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2pkh(script_pub_key[:2] + b"\x40" + script_pub_key[3:])

    # invalid size: 11 bytes instead of 20
    err_msg = "invalid size: "
    with pytest.raises(BTClibValueError, match=err_msg):
        script_pub_key_from_payload(script_type, "00" * 11)


def test_p2wpkh() -> None:

    # self-consistency
    pub_key = "02 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
    payload = hash160(pub_key)
    script_pub_key = script.serialize([0, payload])
    assert_p2wpkh(script_pub_key)
    assert script_pub_key == p2wpkh(pub_key)

    # to the script_pub_key in two steps (through payload)
    script_type = "p2wpkh"
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)

    # back from the script_pub_key to the payload
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)

    # bech32 address
    network = "mainnet"
    address = bech32_address.p2wpkh(pub_key, network)
    assert address == address_from_script_pub_key(script_pub_key, network)
    wit_ver = 0
    assert address == bech32_address_from_witness(wit_ver, payload, network)

    # back from the address to the script_pub_key
    assert (script_pub_key, network) == script_pub_key_from_address(address)

    # p2sh-wrapped base58 address
    address = base58_address.p2wpkh_p2sh(pub_key, network)
    assert address == base58_address_from_witness(payload, network)

    err_msg = "invalid witness version: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2wpkh(b"\x33" + script_pub_key[1:])

    err_msg = "invalid pub_key hash length marker: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2wpkh(script_pub_key[:1] + b"\x00" + script_pub_key[2:])


def test_p2sh() -> None:

    # self-consistency
    pub_key = "02 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
    pub_key_hash = hash160(pub_key)
    redeem_script = script_pub_key_from_payload("p2pkh", pub_key_hash)
    payload = hash160(redeem_script)
    script_pub_key = script.serialize(["OP_HASH160", payload, "OP_EQUAL"])
    assert_p2sh(script_pub_key)
    assert script_pub_key == p2sh(redeem_script)

    # to the script_pub_key in two steps (through payload)
    script_type = "p2sh"
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)

    # back from the script_pub_key to the payload
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)

    # base58 address
    network = "mainnet"
    address = base58_address.p2sh(redeem_script, network)
    assert address == address_from_script_pub_key(script_pub_key, network)
    prefix = NETWORKS[network].p2sh
    assert address == base58_address_from_h160(prefix, payload, network)

    # back from the address to the script_pub_key
    assert (script_pub_key, network) == script_pub_key_from_address(address)

    # documented test case: https://learnmeabitcoin.com/guide/p2sh
    payload = bytes.fromhex("748284390f9e263a4b766a75d0633c50426eb875")
    script_pub_key = bytes.fromhex("a914") + payload + bytes.fromhex("87")
    assert_p2sh(script_pub_key)
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)
    address = "3CK4fEwbMP7heJarmU4eqA3sMbVJyEnU3V"
    assert address == address_from_script_pub_key(script_pub_key, network)
    assert (script_pub_key, network) == script_pub_key_from_address(address)

    err_msg = "missing final OP_EQUAL"
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2sh(script_pub_key[:-1] + b"\x40")

    err_msg = "missing leading OP_HASH160"
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2sh(b"\x40" + script_pub_key[1:])

    err_msg = "invalid redeem script hash length marker: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2sh(script_pub_key[:1] + b"\x40" + script_pub_key[2:])

    # invalid size: 21 bytes instead of 20
    err_msg = "invalid size: "
    with pytest.raises(BTClibValueError, match=err_msg):
        script_pub_key_from_payload(script_type, "00" * 21)


def test_p2wsh() -> None:

    # self-consistency
    pub_key = "02 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
    pub_key_hash = hash160(pub_key)
    redeem_script = script_pub_key_from_payload("p2pkh", pub_key_hash)
    payload = sha256(redeem_script)
    script_pub_key = script.serialize([0, payload])
    assert_p2wsh(script_pub_key)
    assert script_pub_key == p2wsh(redeem_script)

    script_type = "p2wsh"
    assert script_pub_key == script_pub_key_from_payload(script_type, payload)

    # back from the script_pub_key to the payload
    assert (script_type, payload) == payload_from_script_pub_key(script_pub_key)

    # bech32 address
    network = "mainnet"
    address = bech32_address.p2wsh(redeem_script, network)
    assert address == address_from_script_pub_key(script_pub_key, network)
    wit_ver = 0
    assert address == bech32_address_from_witness(wit_ver, payload, network)

    # back from the address to the script_pub_key
    assert (script_pub_key, network) == script_pub_key_from_address(address)

    # p2sh-wrapped base58 address
    address = base58_address.p2wsh_p2sh(redeem_script, network)
    assert address == base58_address_from_witness(payload, network)

    err_msg = "invalid witness version: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2wsh(b"\x33" + script_pub_key[1:])

    err_msg = "invalid redeem script hash length marker: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2wsh(script_pub_key[:1] + b"\x00" + script_pub_key[2:])


def test_unknown() -> None:

    script_pub_key = script.serialize([16, 20 * b"\x00"])
    assert address_from_script_pub_key(script_pub_key) == ""
    assert payload_from_script_pub_key(script_pub_key) == ("unknown", script_pub_key)


def test_exceptions() -> None:

    # invalid size: 11 bytes instead of 20
    err_msg = "invalid size: "
    with pytest.raises(BTClibValueError, match=err_msg):
        script_pub_key_from_payload("p2wpkh", "00" * 11)

    # invalid size: 33 bytes instead of 32
    with pytest.raises(BTClibValueError, match=err_msg):
        script_pub_key_from_payload("p2wsh", "00" * 33)

    err_msg = "unknown script_pub_key type: "
    with pytest.raises(BTClibValueError, match=err_msg):
        script_pub_key_from_payload("p2unkn", "00" * 32)

    # Unhandled witness version (16)
    err_msg = "unmanaged witness version: "
    address = bech32_address_from_witness(16, 20 * b"\x00")
    with pytest.raises(BTClibValueError, match=err_msg):
        script_pub_key_from_address(address)


def test_p2ms_1() -> None:

    # self-consistency
    pub_key0 = "04 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
    pub_key1 = "04 61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765 19aecc62770f5b0e4ef8551946d8a540911abe3e7854a26f39f58b25c15342af"

    # documented test case: https://learnmeabitcoin.com/guide/p2ms
    script_pub_key = bytes.fromhex(  # fmt: off
        "51"  # OP_1
        "41"  # canonical 65-bytes push
        + pub_key0
        + "41"  # noqa E148  # canonical 65-bytes push
        + pub_key1
        + "52"  # noqa E148  # OP_2
        "ae"  # OP_CHECKMULTISIG
    )  # fmt: on
    assert is_p2ms(script_pub_key)
    assert address_from_script_pub_key(script_pub_key) == ""
    script_type, payload = payload_from_script_pub_key(script_pub_key)
    assert script_type == "p2ms"
    assert payload == script_pub_key[:-1]
    assert script_pub_key == script_pub_key_from_payload("p2ms", payload)

    pub_keys: List[Key] = [pub_key0, pub_key1]
    err_msg = "invalid p2ms payload"
    with pytest.raises(BTClibValueError, match=err_msg):
        p2ms(4, pub_keys)
    with pytest.raises(BTClibValueError, match=err_msg):
        p2ms(4, [pub_key0] * 17)
    with pytest.raises(BTClibValueError, match=err_msg):
        p2ms(0, pub_keys)
    with pytest.raises(BTClibValueError, match=err_msg):
        p2ms(17, pub_keys)

    err_msg = "not a private or public key: "
    with pytest.raises(BTClibValueError, match=err_msg):
        p2ms(1, [pub_key0 + "00", pub_key1])

    script_: Script = [1, pub_key0 + "00", pub_key1, 2, "OP_CHECKMULTISIG"]
    script_pub_key = script.serialize(script_)
    assert not is_p2ms(script_pub_key)

    err_msg = "invalid key in p2ms"
    script_pub_key = script.serialize([1, pub_key0, "00", 2, "OP_CHECKMULTISIG"])
    assert not is_p2ms(script_pub_key)

    script_pub_key = script.serialize([1, pub_key0, pub_key1, 2, "OP_CHECKMULTISIG"])
    assert is_p2ms(script_pub_key)

    script_pub_key = script.serialize([2, pub_key0, pub_key1, 2, "OP_CHECKMULTISIG"])
    assert is_p2ms(script_pub_key)

    script_pub_key = script.serialize([0, pub_key0, pub_key1, 2, "OP_CHECKMULTISIG"])
    assert not is_p2ms(script_pub_key)

    script_pub_key = script.serialize([3, pub_key0, pub_key1, 2, "OP_CHECKMULTISIG"])
    assert not is_p2ms(script_pub_key)

    script_pub_key = script.serialize([1, 2, "OP_CHECKMULTISIG"])
    assert not is_p2ms(script_pub_key)

    script_pub_key = script.serialize([1, pub_key0, 2, "OP_CHECKMULTISIG"])
    assert not is_p2ms(script_pub_key)

    script_pub_key = script.serialize([1, pub_key0, pub_key1, 3, "OP_CHECKMULTISIG"])
    assert not is_p2ms(script_pub_key)

    pub_key2 = "04 79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798 483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8"
    script_pub_key = script.serialize(
        [1, pub_key0, pub_key1, pub_key2, 3, "OP_CHECKMULTISIG"]
    )
    assert_p2ms(script_pub_key)

    err_msg = "invalid size: "
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2ms(script_pub_key[:133] + b"\x40" + script_pub_key[134:])

    err_msg = "invalid extra data"
    with pytest.raises(BTClibValueError, match=err_msg):
        assert_p2ms(script_pub_key[:-2] + b"\x00" + script_pub_key[-2:])


def test_p2ms_2() -> None:

    m = 1

    # all uncompressed
    pub_key0 = "04 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
    pub_key1 = "04 61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765 19aecc62770f5b0e4ef8551946d8a540911abe3e7854a26f39f58b25c15342af"
    pub_key2 = "04 79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798 483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8"
    uncompressed_pub_keys: List[Key] = [pub_key0, pub_key1, pub_key2]
    # mixed compressed / uncompressed public keys
    pub_key0 = "04 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
    pub_key1 = "03 61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765"
    pub_key2 = "02 79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798"
    mixed_pub_keys: List[Key] = [pub_key0, pub_key1, pub_key2]

    for pub_keys in (uncompressed_pub_keys, mixed_pub_keys):
        for lexi_sort in (True, False):
            script_pub_key = p2ms(m, pub_keys, lexi_sort=lexi_sort)
            assert is_p2ms(script_pub_key)
            assert address_from_script_pub_key(script_pub_key) == ""
            script_type, payload = payload_from_script_pub_key(script_pub_key)
            assert script_type == "p2ms"
            assert payload == script_pub_key[:-1]
            assert script_pub_key == script_pub_key_from_payload("p2ms", payload)


def test_bip67() -> None:
    "BIP67 test vectors https://en.bitcoin.it/wiki/BIP_0067"

    data_folder = path.join(path.dirname(__file__), "test_data")
    filename = path.join(data_folder, "bip67_test_vectors.json")
    with open(filename, "r") as file_:
        # json.dump(test_vectors, f, indent=4)
        test_vectors = json.load(file_)

    m = 2
    for i in test_vectors:
        keys, address = test_vectors[i]

        script_pub_key = p2ms(m, keys, lexi_sort=True)
        assert is_p2ms(script_pub_key)
        assert address_from_script_pub_key(script_pub_key) == ""
        script_type, payload = payload_from_script_pub_key(script_pub_key)
        assert script_type == "p2ms"
        assert payload == script_pub_key[:-1]
        assert script_pub_key == script_pub_key_from_payload("p2ms", payload)

        errmsg = f"Test vector #{i}"
        assert address == base58_address.p2sh(script_pub_key), errmsg


def test_non_standard_script_in_p2wsh() -> None:

    network = "mainnet"

    fed_pub_keys = [b"\x00" * 33, b"\x11" * 33, b"\x22" * 33]
    rec_pub_keys = [b"\x77" * 33, b"\x88" * 33, b"\x99" * 33]
    # fmt: off
    redeem_script = script.serialize(
        [
            "OP_IF",
                2, *fed_pub_keys, 3, "OP_CHECKMULTISIG",  # noqa E131
            "OP_ELSE",
                500, "OP_CHECKLOCKTIMEVERIFY", "OP_DROP",  # noqa E131
                2, *rec_pub_keys, 3, "OP_CHECKMULTISIG",  # noqa E131
            "OP_ENDIF",
        ]
    )
    # fmt: on
    payload = sha256(redeem_script)
    script_pub_key = (
        "00207b5310339c6001f75614daa5083839fa54d46165f6c56025cc54d397a85a5708"
    )
    assert script_pub_key == p2wsh(redeem_script).hex()
    assert script_pub_key == script_pub_key_from_payload("p2wsh", payload).hex()

    address = "bc1q0df3qvuuvqqlw4s5m2jsswpelf2dgct97mzkqfwv2nfe02z62uyq7n4zjj"
    assert address == address_from_script_pub_key(script_pub_key, network)
    assert address == bech32_address_from_witness(0, payload, network)
